import uuid
import warnings

from google.cloud.exceptions import NotFound

from prefect import context
from prefect.client import Secret
from prefect.core import Task
from prefect.utilities.gcp import get_storage_client
from prefect.utilities.tasks import defaults_from_attrs


class GCSBaseTask(Task):
    def __init__(
        self,
        bucket: str = None,
        blob: str = None,
        project: str = None,
        create_bucket: bool = False,
        encryption_key_secret: str = None,
        **kwargs
    ):
        self.bucket = bucket
        self.blob = blob
        self.project = project
        self.create_bucket = create_bucket
        if encryption_key_secret is not None:
            warnings.warn(
                "The `encryption_key_secret` argument is deprecated. Use a `Secret` task "
                "to pass the key value at runtime instead.",
                UserWarning,
            )
        self.encryption_key_secret = encryption_key_secret
        super().__init__(**kwargs)

    def _retrieve_bucket(self, client, bucket: str, create_bucket: bool):
        "Retrieves a bucket based on user settings"
        try:
            bucket = client.get_bucket(bucket)
        except NotFound as exc:
            if create_bucket is True:
                self.logger.debug("Bucket {} not found; creating...".format(bucket))
                bucket = client.create_bucket(bucket)
            else:
                raise exc
        return bucket

    def _get_blob(
        self,
        bucket: str,
        blob: str,
        encryption_key: str = None,
        encryption_key_secret: str = None,
    ):
        "Retrieves blob based on user settings."
        if blob is None:
            blob = "prefect-" + context.get("task_run_id", "no-id-" + str(uuid.uuid4()))

        # pull encryption_key if requested
        if encryption_key_secret is not None:
            warnings.warn(
                "The `encryption_key_secret` argument is deprecated. Use a `Secret` task "
                "to pass the credentials value at runtime instead.",
                UserWarning,
            )
            encryption_key = Secret(encryption_key_secret).get()

        return bucket.blob(blob, encryption_key=encryption_key)


class GCSDownload(GCSBaseTask):
    """
    Task template for downloading data from Google Cloud Storage as a string.

    Args:
        - bucket (str): default bucket name to download from
        - blob (str, optional): default blob name to download.
        - project (str, optional): default Google Cloud project to work within.
            If not provided, will be inferred from your Google Cloud credentials
        - encryption_key_secret (str, optional, DEPRECATED): the name of the Prefect Secret
            storing an optional `encryption_key` to be used when downloading the Blob
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor

    Note that the design of this task allows you to initialize a _template_ with default
    settings.  Each inidividual occurence of the task in a Flow can overwrite any of these
    default settings for custom use (for example, if you want to pull different credentials for
    a given Task, or specify the Blob name at runtime).
    """

    def __init__(
        self,
        bucket: str,
        blob: str = None,
        project: str = None,
        encryption_key_secret: str = None,
        **kwargs
    ):
        super().__init__(
            bucket=bucket,
            blob=blob,
            project=project,
            encryption_key_secret=encryption_key_secret,
            **kwargs
        )

    @defaults_from_attrs("blob", "bucket", "project", "encryption_key_secret")
    def run(
        self,
        bucket: str = None,
        blob: str = None,
        project: str = None,
        credentials: dict = None,
        encryption_key: str = None,
        encryption_key_secret: str = None,
    ) -> str:
        """
        Run method for this Task.  Invoked by _calling_ this Task after initialization
        within a Flow context.

        Note that some arguments are required for the task to run, and must be provided
        _either_ at initialization _or_ as arguments.

        Args:
            - bucket (str, optional): the bucket name to upload to
            - blob (str, optional): blob name to download from
            - project (str, optional): Google Cloud project to work within. If not provided
                here or at initialization, will be inferred from your Google Cloud credentials
            - credentials (dict, optional): a JSON document containing Google Cloud
                credentials.  You should provide these at runtime with an upstream Secret task.
                If not provided, Prefect will first check `context` for `GCP_CREDENTIALS` and
                lastly will use default Google client logic.
            - encryption_key (str, optional): an encryption key
            - encryption_key_secret (str, optional, DEPRECATED): the name of the Prefect Secret
                storing an optional `encryption_key` to be used when uploading the Blob

        Returns:
            - str: the data from the blob, as a string

        Raises:
            - google.cloud.exception.NotFound: if `create_bucket=False` and the bucket
                name is not found
            - ValueError: if `blob` name hasn't been provided

        """
        # create client
        client = get_storage_client(project=project, credentials=credentials)

        # retrieve bucket
        bucket = self._retrieve_bucket(
            client=client, bucket=bucket, create_bucket=False
        )

        # identify blob name
        gcs_blob = self._get_blob(
            bucket,
            blob,
            encryption_key=encryption_key,
            encryption_key_secret=encryption_key_secret,
        )
        data = gcs_blob.download_as_string()
        return data


class GCSUpload(GCSBaseTask):
    """
    Task template for uploading data to Google Cloud Storage.  Requires the data already
    be a string.

    Args:
        - bucket (str): default bucket name to upload to
        - blob (str, optional): default blob name to upload to; otherwise a random string
            beginning with `prefect-` and containing the Task Run ID will be used
        - project (str, optional): default Google Cloud project to work within.
            If not provided, will be inferred from your Google Cloud credentials
        - create_bucket (bool, optional): boolean specifying whether to create the bucket if it
            does not exist, otherwise an Exception is raised. Defaults to `False`.
        - encryption_key_secret (str, optional, DEPRECATED): the name of the Prefect Secret
            storing an optional `encryption_key` to be used when uploading the Blob
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor

    Note that the design of this task allows you to initialize a _template_ with default
    settings.  Each inidividual occurence of the task in a Flow can overwrite any of
    these default settings for custom use (for example, if you want to pull different
    credentials for a given Task, or specify the Blob name at runtime).
    """

    def __init__(
        self,
        bucket: str,
        blob: str = None,
        project: str = None,
        create_bucket: bool = False,
        encryption_key_secret: str = None,
        **kwargs
    ):
        super().__init__(
            bucket=bucket,
            blob=blob,
            project=project,
            create_bucket=create_bucket,
            encryption_key_secret=encryption_key_secret,
            **kwargs
        )

    @defaults_from_attrs(
        "bucket", "blob", "project", "create_bucket", "encryption_key_secret",
    )
    def run(
        self,
        data: str,
        bucket: str = None,
        blob: str = None,
        project: str = None,
        credentials: dict = None,
        encryption_key: str = None,
        create_bucket: bool = False,
        encryption_key_secret: str = None,
    ) -> str:
        """
        Run method for this Task.  Invoked by _calling_ this Task after initialization
        within a Flow context.

        Note that some arguments are required for the task to run, and must be
        provided _either_ at initialization _or_ as arguments.

        Args:
            - data (str): the data to upload; must already be represented as a string
            - bucket (str, optional): the bucket name to upload to
            - blob (str, optional): blob name to upload to
                a string beginning with `prefect-` and containing the Task Run ID will be used
            - project (str, optional): Google Cloud project to work within. Can be inferred
                from credentials if not provided.
            - credentials (dict, optional): a JSON document containing Google Cloud credentials.
                You should provide these at runtime with an upstream Secret task.  If not
                provided, Prefect will first check `context` for `GCP_CREDENTIALS` and lastly
                will use default Google client logic.
            - encryption_key (str, optional): an encryption key
            - create_bucket (bool, optional): boolean specifying whether to create the bucket
                if it does not exist, otherwise an Exception is raised. Defaults to `False`.
            - encryption_key_secret (str, optional, DEPRECATED): the name of the Prefect Secret
                storing an optional `encryption_key` to be used when uploading the Blob.

        Raises:
            - google.cloud.exception.NotFound: if `create_bucket=False` and the bucket name is
                not found

        Returns:
            - str: the blob name that now stores the provided data
        """
        # create client
        client = get_storage_client(project=project, credentials=credentials)

        # retrieve bucket
        bucket = self._retrieve_bucket(
            client=client, bucket=bucket, create_bucket=create_bucket
        )

        # identify blob name
        gcs_blob = self._get_blob(
            bucket,
            blob,
            encryption_key=encryption_key,
            encryption_key_secret=encryption_key_secret,
        )
        gcs_blob.upload_from_string(data)
        return gcs_blob.name


class GCSCopy(GCSBaseTask):
    """
    Task template for copying data from one Google Cloud Storage bucket to another, without
    downloading it locally.

    Note that some arguments are required for the task to run, and must be
    provided _either_ at initialization _or_ as arguments.

    Args:
        - source_bucket (str, optional): default source bucket name.
        - source_blob (str, optional): default source blob name.
        - dest_bucket (str, optional): default destination bucket name.
        - dest_blob (str, optional): default destination blob name.
        - project (str, optional): default Google Cloud project to work within.
            If not provided, will be inferred from your Google Cloud credentials
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor

    Note that the design of this task allows you to initialize a _template_ with default
    settings.  Each inidividual occurence of the task in a Flow can overwrite any of these
    default settings for custom use (for example, if you want to pull different credentials for
    a given Task, or specify the Blob name at runtime).
    """

    def __init__(
        self,
        source_bucket: str = None,
        source_blob: str = None,
        dest_bucket: str = None,
        dest_blob: str = None,
        project: str = None,
        **kwargs
    ):
        self.source_bucket = source_bucket
        self.source_blob = source_blob
        self.dest_bucket = dest_bucket
        self.dest_blob = dest_blob

        super().__init__(project=project, **kwargs)

    @defaults_from_attrs(
        "source_bucket", "source_blob", "dest_bucket", "dest_blob", "project",
    )
    def run(
        self,
        source_bucket: str = None,
        source_blob: str = None,
        dest_bucket: str = None,
        dest_blob: str = None,
        project: str = None,
        credentials: dict = None,
    ) -> str:
        """
        Run method for this Task. Invoked by _calling_ this Task after initialization
        within a Flow context.

        Note that some arguments are required for the task to run, and must be
        provided _either_ at initialization _or_ as arguments.

        Args:
            - source_bucket (str, optional): default source bucket name.
            - source_blob (str, optional): default source blob name.
            - dest_bucket (str, optional): default destination bucket name.
            - dest_blob (str, optional): default destination blob name.
            - project (str, optional): default Google Cloud project to work within.
                If not provided, will be inferred from your Google Cloud credentials
            - credentials (dict, optional): a JSON document containing Google Cloud credentials.
                You should provide these at runtime with an upstream Secret task.  If not
                provided, Prefect will first check `context` for `GCP_CREDENTIALS` and lastly
                will use default Google client logic.

        Returns:
            - str: the name of the destination blob

        Raises:
            - ValueError: if `source_bucket`, `source_blob`, `dest_bucket`, or `dest_blob`
                are missing or point at the same object.

        """
        if None in [source_bucket, source_blob, dest_bucket, dest_blob]:
            raise ValueError("Missing source or destination")
        elif (source_bucket, source_blob) == (dest_bucket, dest_blob):
            raise ValueError("Source and destination are identical.")

        # create client
        client = get_storage_client(project=project, credentials=credentials)

        # get source bucket and blob
        source_bucket_obj = client.get_bucket(source_bucket)
        source_blob_obj = source_bucket_obj.blob(source_blob)
        # get dest bucket
        dest_bucket_obj = client.get_bucket(dest_bucket)
        # copy from source blob to dest bucket
        source_bucket_obj.copy_blob(
            blob=source_blob_obj, destination_bucket=dest_bucket_obj, new_name=dest_blob
        )

        return dest_blob



import importlib
from typing import Any


def import_object(name: str) -> Any:
    """Import an object given a fully-qualified name.

    Args:
        - name (string): The fully-qualified name of the object to import.

    Returns:
        - obj: The object that was imported.

    Example:

    ```python
    >>> obj = import_object("random.randint")
    >>> import random
    >>> obj == random.randint
    True
    ```
    """
    try:
        mod_name, attr_name = name.rsplit(".", 1)
        mod = importlib.import_module(mod_name)
        return getattr(mod, attr_name)
    except ValueError:
        return importlib.import_module(name)



import copy
import os
import uuid
import warnings
from typing import Any, Callable, List, TYPE_CHECKING

import yaml

import prefect
from prefect.environments.execution.base import Environment, _RunMixin
from prefect.utilities.storage import get_flow_image

if TYPE_CHECKING:
    from prefect.core.flow import Flow  # pylint: disable=W0611


class KubernetesJobEnvironment(Environment, _RunMixin):
    """
    KubernetesJobEnvironment is an environment which deploys your flow as a Kubernetes
    job. This environment allows (and requires) a custom job YAML spec to be provided.

    When providing a custom YAML job spec the first container in the spec must be the
    container that the flow runner will be executed on.

    The following environment variables, required for cloud, do not need to be
    included––they are automatically added and populated during execution:

    - `PREFECT__CLOUD__GRAPHQL`
    - `PREFECT__CLOUD__AUTH_TOKEN`
    - `PREFECT__CONTEXT__FLOW_RUN_ID`
    - `PREFECT__CONTEXT__NAMESPACE`
    - `PREFECT__CONTEXT__IMAGE`
    - `PREFECT__CLOUD__USE_LOCAL_SECRETS`
    - `PREFECT__ENGINE__FLOW_RUNNER__DEFAULT_CLASS`
    - `PREFECT__ENGINE__TASK_RUNNER__DEFAULT_CLASS`
    - `PREFECT__LOGGING__LOG_TO_CLOUD`
    - `PREFECT__LOGGING__EXTRA_LOGGERS`

    Additionally, the following command will be applied to the first container:
    `$ /bin/sh -c "python -c 'import prefect; prefect.environments.execution.load_and_run_flow()'"`

    Args:
        - job_spec_file (str, optional): Path to a job spec YAML file. This path is only
            used when the environment is built, so should refer to a file on the machine
            used to build the flow.
        - unique_job_name (bool, optional): whether to use a unique name for each job created
            with this environment. Defaults to `False`
        - executor (Executor, optional): the executor to run the flow with. If not provided, the
            default executor will be used.
        - executor_kwargs (dict, optional): DEPRECATED
        - labels (List[str], optional): a list of labels, which are arbitrary string
            identifiers used by Prefect Agents when polling for work
        - on_start (Callable, optional): a function callback which will be called before the
            flow begins to run
        - on_exit (Callable, optional): a function callback which will be called after the flow
            finishes its run
        - metadata (dict, optional): extra metadata to be set and serialized on this environment
    """

    def __init__(
        self,
        job_spec_file: str = None,
        unique_job_name: bool = False,
        executor: "prefect.engine.executors.Executor" = None,
        executor_kwargs: dict = None,
        labels: List[str] = None,
        on_start: Callable = None,
        on_exit: Callable = None,
        metadata: dict = None,
    ) -> None:
        self.job_spec_file = os.path.abspath(job_spec_file) if job_spec_file else None
        self.unique_job_name = unique_job_name

        if executor_kwargs is not None:
            warnings.warn("`executor_kwargs` is deprecated, use `executor` instead")
        if executor is None:
            executor = prefect.engine.get_default_executor_class()(
                **(executor_kwargs or {})
            )
        elif not isinstance(executor, prefect.engine.executors.Executor):
            raise TypeError(
                f"`executor` must be an `Executor` or `None`, got `{executor}`"
            )
        self.executor = executor

        # Load specs from file if path given, store on object
        self._job_spec = self._load_spec_from_file()
        self._job_spec = self._populate_build_time_job_spec_details(self._job_spec)

        self._identifier_label = ""

        super().__init__(
            labels=labels, on_start=on_start, on_exit=on_exit, metadata=metadata
        )

    @property
    def dependencies(self) -> list:
        return ["kubernetes"]

    @property
    def identifier_label(self) -> str:
        if not hasattr(self, "_identifier_label") or not self._identifier_label:
            self._identifier_label = str(uuid.uuid4())
        return self._identifier_label

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        # Ensure _identifier_label is not persisted
        if "_identifier_label" in state:
            del state["_identifier_label"]
        return state

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)

    def execute(  # type: ignore
        self, flow: "Flow", **kwargs: Any
    ) -> None:
        """
        Create a single Kubernetes job that runs the flow.

        Args:
            - flow (Flow): the Flow object
            - **kwargs (Any): additional keyword arguments to pass to the runner

        Raises:
            - Exception: if the environment is unable to create the Kubernetes job
        """
        docker_name = get_flow_image(flow)

        from kubernetes import client, config

        # Verify environment is running in cluster
        try:
            config.load_incluster_config()
        except config.config_exception.ConfigException:
            self.logger.error("Environment not currently running inside a cluster")
            raise EnvironmentError("Environment not currently inside a cluster")

        batch_client = client.BatchV1Api()

        job = self._populate_run_time_job_spec_details(docker_name=docker_name,)

        # Create Job
        try:
            batch_client.create_namespaced_job(
                namespace=prefect.context.get("namespace"), body=job
            )
        except Exception as exc:
            self.logger.critical("Failed to create Kubernetes job: {}".format(exc))
            raise exc

    ###############################
    # Custom YAML Spec Manipulation
    ###############################

    @staticmethod
    def _ensure_required_job_spec_sections(yaml_obj: dict) -> dict:
        """
        Ensure that the required sections exist in the given job YAML.

        Makes sure the following sections exist:

        * `metadata`
        * `metadata.labels`
        * `spec`
        * `spec.template`
        * `spec.template.metadata`
        * `spec.template.metadata.labels`
        * `spec.template.spec`
        * `spec.template.spec.containers`
        * and on the first container in `spec.template.spec.containers`:
            - `command`
            - `args`

        Args:
            - yaml_obj (dict): A dictionary representing the parsed yaml

        Returns:
            - dict: a dictionary with the yaml values replaced
        """
        if not yaml_obj.get("metadata"):
            yaml_obj["metadata"] = {}

        if not yaml_obj["metadata"].get("labels"):
            yaml_obj["metadata"]["labels"] = {}

        if not yaml_obj.get("spec"):
            yaml_obj["spec"] = {}

        if not yaml_obj["spec"].get("template"):
            yaml_obj["spec"]["template"] = {}

        if not yaml_obj["spec"]["template"].get("spec"):
            yaml_obj["spec"]["template"]["spec"] = {}

        if not yaml_obj["spec"]["template"]["spec"].get("containers"):
            yaml_obj["spec"]["template"]["spec"]["containers"] = {}

        if not yaml_obj["spec"]["template"].get("metadata"):
            yaml_obj["spec"]["template"]["metadata"] = {}

        if not yaml_obj["spec"]["template"]["metadata"].get("labels"):
            yaml_obj["spec"]["template"]["metadata"]["labels"] = {}

        if not yaml_obj["spec"]["template"]["spec"].get("containers"):
            yaml_obj["spec"]["template"]["spec"]["containers"] = [{}]

        if not yaml_obj["spec"]["template"]["spec"]["containers"][0].get("command"):
            yaml_obj["spec"]["template"]["spec"]["containers"][0]["command"] = []

        if not yaml_obj["spec"]["template"]["spec"]["containers"][0].get("args"):
            yaml_obj["spec"]["template"]["spec"]["containers"][0]["args"] = []

        return yaml_obj

    def _populate_build_time_job_spec_details(self, yaml_obj: dict) -> dict:
        """
        Populate some details of the custom execution job YAML used in this environment.

        This method fills in details that are known at the build time (when assigning
        `flow.environment`). Other details which can only be filled in at runtime are
        handled by `_populate_job_spec_yaml()`.

        Changes the first container in `spec.template.spec.containers`.

        * `/bin/sh -c` as the `command`
        * prefect-specific `args` that run the floow

        Args:
            - yaml_obj (dict): A dictionary representing the parsed yaml

        Returns:
            - dict: a dictionary with the yaml values replaced
        """
        yaml_obj = self._ensure_required_job_spec_sections(yaml_obj)

        # set command on first container
        yaml_obj["spec"]["template"]["spec"]["containers"][0]["command"] = [
            "/bin/sh",
            "-c",
        ]

        # set args on first container
        yaml_obj["spec"]["template"]["spec"]["containers"][0]["args"] = [
            'python -c "import prefect; prefect.environments.execution.load_and_run_flow()"'
        ]

        return yaml_obj

    def _populate_run_time_job_spec_details(self, docker_name: str) -> dict:
        """
        Fill in the custom execution job yaml object stored in `self._job_spec`
        with relevant details.

        * `metadata.name`: adds a random name if `self.unique_job_name` is True
        * `metadata.labels`: appends prefect-specific labels
        * `spec.template.metadata.labels`: appends prefect-specific labels
        * `spec.template.spec.containers` (first container):
            - `env`: appends prefect-specific environment variables
            - `image`: writes in image from flow's storage or evironment metadata

        Args:
            - docker_name (str): the full path to the docker image

        Returns:
            - dict: a dictionary with the yaml values replaced
        """
        flow_run_id = prefect.context.get("flow_run_id", "unknown")

        yaml_obj = copy.deepcopy(self._job_spec)
        yaml_obj = self._ensure_required_job_spec_sections(yaml_obj)

        if self.unique_job_name:
            yaml_obj["metadata"][
                "name"
            ] = f"{yaml_obj['metadata']['name']}-{str(uuid.uuid4())[:8]}"

        # Populate metadata label fields
        k8s_labels = {
            "prefect.io/identifier": self.identifier_label,
            "prefect.io/flow_run_id": flow_run_id,
        }
        yaml_obj["metadata"]["labels"].update(k8s_labels)
        yaml_obj["spec"]["template"]["metadata"]["labels"].update(k8s_labels)

        # Required Cloud environment variables
        env_values = [
            {"name": "PREFECT__CLOUD__GRAPHQL", "value": prefect.config.cloud.graphql},
            {
                "name": "PREFECT__CLOUD__AUTH_TOKEN",
                "value": prefect.config.cloud.auth_token,
            },
            {"name": "PREFECT__CONTEXT__FLOW_RUN_ID", "value": flow_run_id},
            {
                "name": "PREFECT__CONTEXT__NAMESPACE",
                "value": prefect.context.get("namespace", ""),
            },
            {"name": "PREFECT__CONTEXT__IMAGE", "value": docker_name},
            {"name": "PREFECT__CLOUD__USE_LOCAL_SECRETS", "value": "false"},
            {
                "name": "PREFECT__ENGINE__FLOW_RUNNER__DEFAULT_CLASS",
                "value": "prefect.engine.cloud.CloudFlowRunner",
            },
            {
                "name": "PREFECT__ENGINE__TASK_RUNNER__DEFAULT_CLASS",
                "value": "prefect.engine.cloud.CloudTaskRunner",
            },
            {"name": "PREFECT__LOGGING__LOG_TO_CLOUD", "value": "true"},
            {
                "name": "PREFECT__LOGGING__EXTRA_LOGGERS",
                "value": str(prefect.config.logging.extra_loggers),
            },
        ]

        # set environment variables on all containers
        for container in yaml_obj["spec"]["template"]["spec"]["containers"]:
            if not container.get("env"):
                container["env"] = []
            container["env"].extend(env_values)

        # set image on first container
        yaml_obj["spec"]["template"]["spec"]["containers"][0]["image"] = docker_name

        return yaml_obj

    def _load_spec_from_file(self) -> dict:
        """
        Load job spec from provided file path

        Returns:
            - dict: job spec dictionary
        """
        job = dict()  # type: ignore

        if self.job_spec_file:
            with open(self.job_spec_file) as job_spec_file:
                job = yaml.safe_load(job_spec_file)

        return job



"""
Tasks that interface with various components of Google Cloud Platform.

Note that these tasks allow for a wide range of custom usage patterns, such as:

- Initialize a task with all settings for one time use
- Initialize a "template" task with default settings and override as needed
- Create a custom Task that inherits from a Prefect Task and utilizes the Prefect boilerplate
"""
import warnings

warnings.warn(
    "DEPRECATED: prefect.tasks.google is deprecated, use prefect.tasks.gcp instead",
    UserWarning,
)

from prefect.tasks.gcp.storage import GCSDownload, GCSUpload, GCSCopy
from prefect.tasks.gcp.bigquery import (
    BigQueryTask,
    BigQueryLoadGoogleCloudStorage,
    BigQueryStreamingInsert,
    CreateBigQueryTable,
)



"""
Environments are JSON-serializable objects that fully describe how to run a flow. Serialization
schemas are contained in `prefect.serialization.environment.py`.

Different Environment objects correspond to different computation environments. Environments
that are written on top of a type of infrastructure also define how to set up and execute
that environment. e.g. the `DaskKubernetesEnvironment` is an environment which
runs a flow on Kubernetes using the `dask-kubernetes` library.
"""

from typing import Callable, Iterable, TYPE_CHECKING

import prefect
from prefect.client import Client
from prefect.utilities import logging
from prefect.utilities.graphql import with_args

if TYPE_CHECKING:
    from prefect.core.flow import Flow  # pylint: disable=W0611


class Environment:
    """
    Base class for Environments.

    An environment is an object that can be instantiated in a way that makes it possible to
    call `environment.setup()` to stand up any required static infrastructure and
    `environment.execute()` to execute the flow inside this environment.

    The `setup` and `execute` functions of an environment require a Prefect Flow object.

    Args:
        - labels (List[str], optional): a list of labels, which are arbitrary string
            identifiers used by Prefect Agents when polling for work
        - on_start (Callable, optional): a function callback which will be called before the
            flow begins to run
        - on_exit (Callable, optional): a function callback which will be called after the flow
            finishes its run
        - metadata (dict, optional): extra metadata to be set and serialized on this environment
    """

    def __init__(
        self,
        labels: Iterable[str] = None,
        on_start: Callable = None,
        on_exit: Callable = None,
        metadata: dict = None,
    ) -> None:
        self.labels = set(labels) if labels else set()
        self.on_start = on_start
        self.on_exit = on_exit
        self.metadata = metadata or {}
        self.logger = logging.get_logger(type(self).__name__)

    def __repr__(self) -> str:
        return "<Environment: {}>".format(type(self).__name__)

    @property
    def dependencies(self) -> list:
        return []

    def setup(self, flow: "Flow") -> None:
        """
        Sets up any infrastructure needed for this environment

        Args:
            - flow (Flow): the Flow object
        """

    def execute(self, flow: "Flow") -> None:
        """
        Executes the flow for this environment.

        Args:
            - flow (Flow): the Flow object
        """

    def serialize(self) -> dict:
        """
        Returns a serialized version of the Environment

        Returns:
            - dict: the serialized Environment
        """
        schema = prefect.serialization.environment.EnvironmentSchema()
        return schema.dump(self)


class _RunMixin:
    """This mixin will go away when all environments share the same run
    implementation.

    For now this is just to share code between a few of the environments"""

    def run(self, flow: "Flow") -> None:
        """
        Run the flow using this environment.

        Args:
            - flow (Flow): the flow object
        """
        assert isinstance(self, Environment)  # mypy
        if self.on_start:
            self.on_start()

        try:
            from prefect.engine import get_default_flow_runner_class

            runner_cls = get_default_flow_runner_class()
            runner_cls(flow=flow).run(executor=self.executor)  # type: ignore
        except Exception as exc:
            self.logger.exception(
                "Unexpected error raised during flow run: {}".format(exc)
            )
            raise exc
        finally:
            if self.on_exit:
                self.on_exit()


def load_and_run_flow() -> None:
    """
    Loads a flow (and the corresponding environment), then runs the flow with
    the environment.

    This is useful for environments whose `execute` method schedules a job that
    later needs to run the flow.

    Raises:
        - ValueError: if no `flow_run_id` is found in context
    """
    logger = logging.get_logger("Environment")
    try:
        flow_run_id = prefect.context.get("flow_run_id")

        if not flow_run_id:
            raise ValueError("No flow run ID found in context.")

        query = {
            "query": {
                with_args("flow_run", {"where": {"id": {"_eq": flow_run_id}}}): {
                    "flow": {"name": True, "storage": True},
                }
            }
        }

        client = Client()
        result = client.graphql(query)
        flow_run = result.data.flow_run[0]

        flow_data = flow_run.flow
        storage_schema = prefect.serialization.storage.StorageSchema()
        storage = storage_schema.load(flow_data.storage)

        # populate global secrets
        secrets = prefect.context.get("secrets", {})
        for secret in storage.secrets:
            secrets[secret] = prefect.tasks.secrets.PrefectSecret(name=secret).run()

        with prefect.context(secrets=secrets):
            flow = storage.get_flow(storage.flows[flow_data.name])
            flow.environment.run(flow)
    except Exception as exc:
        logger.exception("Unexpected error raised during flow run: {}".format(exc))
        raise exc



from typing import Any, Callable, Dict

from toolz import curry

import prefect
from prefect import Task, Flow
from prefect.core import Edge
from prefect.engine.state import State
from prefect.engine import signals


__all__ = ("resource_manager", "ResourceManager")


class ResourceSetupTask(Task):
    """Setup a resource with its resource manager"""

    def run(self, mgr: Any) -> Any:
        return mgr.setup()


class ResourceCleanupTask(Task):
    """Cleanup a resource with its resource manager"""

    def run(self, mgr: Any, resource: Any) -> None:
        mgr.cleanup(resource)


def resource_cleanup_trigger(upstream_states: Dict[Edge, State]) -> bool:
    """Run the cleanup task, provided the following hold:

    - All upstream tasks have finished
    - The resource init task succeeded and wasn't skipped
    - The resource setup task succeeded and wasn't skipped
    """
    for edge, state in upstream_states.items():
        if not state.is_finished():
            raise signals.TRIGGERFAIL(
                "Trigger was 'resource_cleanup_trigger' but some of the "
                "upstream tasks were not finished."
            )
        if edge.key == "mgr":
            if state.is_skipped():
                raise signals.SKIP("Resource manager init skipped")
            elif not state.is_successful():
                raise signals.SKIP("Resource manager init failed")
        elif edge.key == "resource":
            if state.is_skipped():
                raise signals.SKIP("Resource manager setup skipped")
            elif not state.is_successful():
                raise signals.SKIP("Resource manager setup failed")
    return True


class ResourceContext:
    """A context managed by a `ResourceManager`.

    These objects usually are created by calling a `ResourceManager` object,
    not manually. See the docstrings for `ResourceManager` or
    `resource_manager` for more information.
    """

    def __init__(
        self, init_task: Task, setup_task: Task, cleanup_task: Task, flow: Flow,
    ):
        self.init_task = init_task
        self.setup_task = setup_task
        self.cleanup_task = cleanup_task
        self._flow = flow
        self._tasks = set()

    def add_task(self, task: Task, flow: Flow) -> None:
        """Add a new task under the resource manager block.

        Args:
            - task (Task): the task to add
            - flow (Flow): the flow to use
        """
        if self._flow is not flow:
            raise ValueError(
                "Multiple flows cannot be used with the same resource block"
            )
        self._tasks.add(task)

    def __enter__(self):
        self.__prev_resource = prefect.context.get("resource")
        prefect.context.update(resource=self)
        return self.setup_task

    def __exit__(self, *args):
        if self.__prev_resource is None:
            prefect.context.pop("resource", None)
        else:
            prefect.context.update(resource=self.__prev_resource)

        for child in self._tasks:
            # If a task has no upstream tasks created in this resource block,
            # the resource setup should be set as an upstream task.
            # Likewise, if a task has no downstream tasks created in this resource block,
            # the resource cleanup should be set as a downstream task.
            upstream = self._flow.upstream_tasks(child)
            if (
                not self._tasks.intersection(upstream)
                and self.setup_task not in upstream
            ):
                child.set_upstream(self.setup_task, flow=self._flow)
            downstream = self._flow.downstream_tasks(child)
            if (
                not self._tasks.intersection(downstream)
                and self.cleanup_task not in downstream
            ):
                child.set_downstream(self.cleanup_task, flow=self._flow)


class ResourceManager:
    """An object for managing temporary resources.

    Used as a context manager, `ResourceManager` objects create tasks to setup
    and cleanup temporary objects used within a block of tasks.  Examples might
    include temporary Dask/Spark clusters, Docker containers, etc...

    `ResourceManager` objects are usually created using the `resource_manager`
    decorator, but can be created directly using this class if desired.

    For more information, see the docs for `resource_manager`.

    Args:
        - resource_class (Callable): A callable (usually the class itself) for
            creating an object that follows the `ResourceManager` protocol.
        - name (str, optional): The resource name - defaults to the name of the
            decorated class.
        - init_task_kwargs (dict, optional): keyword arguments that will be
            passed to the `Task` constructor for the `init` task.
        - setup_task_kwargs (dict, optional): keyword arguments that will be
            passed to the `Task` constructor for the `setup` task.
        - cleanup_task_kwargs (dict, optional): keyword arguments that will be
            passed to the `Task` constructor for the `cleanup` task.

    Example:

    Here's an example resource manager for creating a temporary local dask
    cluster as part of the flow.

    ```python
    from prefect import resource_manager
    from dask.distributed import Client

    @resource_manager
    class DaskCluster:
        def __init__(self, n_workers):
            self.n_workers = n_workers

        def setup(self):
            "Create a local dask cluster"
            return Client(n_workers=self.n_workers)

        def cleanup(self, client):
            "Cleanup the local dask cluster"
            client.close()
    ```

    To use the `DaskCluster` resource manager as part of your Flow, you can use
    `DaskCluster` as a context manager:

    ```python
    with Flow("example") as flow:
        n_workers = Parameter("n_workers")

        with DaskCluster(n_workers=n_workers) as client:
            some_task(client)
            some_other_task(client)
    ```
    """

    def __init__(
        self,
        resource_class: Callable,
        name: str = None,
        init_task_kwargs: dict = None,
        setup_task_kwargs: dict = None,
        cleanup_task_kwargs: dict = None,
    ):
        self.resource_class = resource_class
        self.init_task_kwargs = (init_task_kwargs or {}).copy()
        self.setup_task_kwargs = (setup_task_kwargs or {}).copy()
        self.cleanup_task_kwargs = (cleanup_task_kwargs or {}).copy()

        if name is None:
            name = getattr(resource_class, "__name__", "resource")
        self.name = name
        self.init_task_kwargs.setdefault("name", name)
        self.setup_task_kwargs.setdefault("name", f"{name}.setup")
        self.cleanup_task_kwargs.setdefault("name", f"{name}.cleanup")
        self.cleanup_task_kwargs.setdefault("trigger", resource_cleanup_trigger)
        self.cleanup_task_kwargs.setdefault("skip_on_upstream_skip", False)

    def __call__(self, *args: Any, flow: Flow = None, **kwargs: Any) -> ResourceContext:
        if flow is None:
            flow = prefect.context.get("flow")
            if flow is None:
                raise ValueError("Could not infer an active Flow context.")

        init_task = prefect.task(self.resource_class, **self.init_task_kwargs)(
            *args, flow=flow, **kwargs
        )

        setup_task = ResourceSetupTask(**self.setup_task_kwargs)(init_task, flow=flow)

        cleanup_task = ResourceCleanupTask(**self.cleanup_task_kwargs)(
            init_task, setup_task, flow=flow
        )

        return ResourceContext(init_task, setup_task, cleanup_task, flow)


@curry
def resource_manager(
    resource_class: Callable,
    name: str = None,
    init_task_kwargs: dict = None,
    setup_task_kwargs: dict = None,
    cleanup_task_kwargs: dict = None,
) -> ResourceManager:
    """A decorator for creating a `ResourceManager` object.

    Used as a context manager, `ResourceManager` objects create tasks to setup
    and cleanup temporary objects used within a block of tasks.  Examples might
    include temporary Dask/Spark clusters, Docker containers, etc...

    Through usage a ResourceManager object adds three tasks to the graph:
        - A `init` task, which returns an object that meets the `ResourceManager`
          protocol. This protocol requires two methods:
            * `setup(self) -> resource`: A method for creating the resource.
                The return value from this will available to user tasks.
            * `cleanup(self, resource) -> None`: A method for cleaning up the
                resource.  This takes the return value from `setup` and
                shouldn't return anything.
        - A `setup` task, which calls the `setup` method on the `ResourceManager`
        - A `cleanup` task, which calls the `cleanup` method on the `ResourceManager`.

    Args:
        - resource_class (Callable): The decorated class.
        - name (str, optional): The resource name - defaults to the name of the
            decorated class.
        - init_task_kwargs (dict, optional): keyword arguments that will be
            passed to the `Task` constructor for the `init` task.
        - setup_task_kwargs (dict, optional): keyword arguments that will be
            passed to the `Task` constructor for the `setup` task.
        - cleanup_task_kwargs (dict, optional): keyword arguments that will be
            passed to the `Task` constructor for the `cleanup` task.

    Returns:
        - ResourceManager: the created `ResourceManager` object.

    Example:

    Here's an example resource manager for creating a temporary local dask
    cluster as part of the flow.

    ```
    from prefect import resource_manager
    from dask.distributed import Client

    @resource_manager
    class DaskCluster:
        def __init__(self, n_workers):
            self.n_workers = n_workers

        def setup(self):
            "Create a local dask cluster"
            return Client(n_workers=self.n_workers)

        def cleanup(self, client):
            "Cleanup the local dask cluster"
            client.close()
    ```

    To use the `DaskCluster` resource manager as part of your Flow, you can use
    `DaskCluster` as a context manager:

    ```
    with Flow("example") as flow:
        n_workers = Parameter("n_workers")

        with DaskCluster(n_workers=n_workers) as client:
            some_task(client)
            some_other_task(client)
    ```

    The `Task` returned by entering the `DaskCluster` context (i.e. the
    `client` part of  `as client`) is the output of the `setup` method on the
    `ResourceManager` class. A `Task` is automatically added to call the
    `cleanup` method (closing the Dask cluster) after all tasks under the
    context have completed. By default this `cleanup` task is configured with
    a trigger to always run if the `setup` task succeeds, and won't be set as
    a reference task.
    """
    return ResourceManager(
        resource_class,
        name=name,
        init_task_kwargs=init_task_kwargs,
        setup_task_kwargs=setup_task_kwargs,
        cleanup_task_kwargs=cleanup_task_kwargs,
    )



import prefect.utilities.logging
import prefect.utilities.backend
import prefect.utilities.collections
import prefect.utilities.configuration
import prefect.utilities.datetimes
import prefect.utilities.diagnostics
import prefect.utilities.exceptions
import prefect.utilities.graphql
import prefect.utilities.notifications
import prefect.utilities.serialization
import prefect.utilities.tasks
import prefect.utilities.agent
import prefect.utilities.storage



import warnings
from typing import Callable, List

from distributed.security import Security

from prefect.environments.execution.remote import RemoteEnvironment


class RemoteDaskEnvironment(RemoteEnvironment):
    """
    RemoteDaskEnvironment is an environment which takes the address of an existing
    Dask cluster and runs the flow on that cluster using `DaskExecutor`.

    Example:
    ```python
    # using a RemoteDaskEnvironment with an existing Dask cluster

    env = RemoteDaskEnvironment(
        address="tcp://dask_scheduler_host_or_ip:8786"
    )

    f = Flow("dummy flow", environment=env)
    ```

    If using the Security object then it should be filled out like this:

    ```python
    security = Security(tls_ca_file='cluster_ca.pem',
                        tls_client_cert='cli_cert.pem',
                        tls_client_key='cli_key.pem',
                        require_encryption=True)
    ```

    For more on using TLS with Dask see https://distributed.dask.org/en/latest/tls.html


    Args:
        - address (str): an address of the scheduler of a Dask cluster in URL form,
            e.g. `tcp://172.33.17.28:8786`
        - security (Security, optional): a Dask Security object from
          `distributed.security.Security`.  Use this to connect to a Dask cluster that is
          enabled with TLS encryption.
        - executor_kwargs (dict, optional): a dictionary of kwargs to be passed to
            the executor; defaults to an empty dictionary
        - labels (List[str], optional): a list of labels, which are arbitrary string
            identifiers used by Prefect Agents when polling for work
        - on_start (Callable, optional): a function callback which will be called before the
            flow begins to run
        - on_exit (Callable, optional): a function callback which will be called after the flow
            finishes its run
        - metadata (dict, optional): extra metadata to be set and serialized on this environment
    """

    def __init__(
        self,
        address: str,
        security: Security = None,
        executor_kwargs: dict = None,
        labels: List[str] = None,
        on_start: Callable = None,
        on_exit: Callable = None,
        metadata: dict = None,
    ) -> None:
        if type(self) is RemoteDaskEnvironment or not type(self).__module__.startswith(
            "prefect."
        ):
            # Only warn if its a subclass not part of prefect, since we don't
            # want to update the code for e.g. `DaskCloudProviderEnvironment`
            warnings.warn(
                "`RemoteDaskEnvironment` is deprecated, please use `LocalEnvironment` with a "
                "`DaskExecutor` instead."
            )
        self.address = address
        dask_executor_kwargs = executor_kwargs or dict()
        dask_executor_kwargs["address"] = address

        if security:
            dask_executor_kwargs["security"] = security

        super().__init__(
            executor="prefect.engine.executors.DaskExecutor",
            executor_kwargs=dask_executor_kwargs,
            labels=labels,
            on_start=on_start,
            on_exit=on_exit,
            metadata=metadata,
        )

    @property
    def dependencies(self) -> list:
        return []



import prefect.schedules.clocks
import prefect.schedules.filters
import prefect.schedules.adjustments
import prefect.schedules.schedules
from prefect.schedules.schedules import (
    Schedule,
    IntervalSchedule,
    CronSchedule,
    OneTimeSchedule,
    UnionSchedule,
)



"""
Note: Result Handlers have been deprecated in 0.11.0.

Result Handlers provide the hooks that Prefect uses to store task results in production for
Prefect < 0.11.0; a `ResultHandler` can be provided to a `Flow` at creation.

Anytime a task needs its output or inputs stored, a result handler is used to determine where
this data should be stored (and how it can be retrieved).
"""
from typing import Any

from prefect.utilities import logging


class ResultHandler:
    def __init__(self) -> None:
        self.logger = logging.get_logger(type(self).__name__)

    def __repr__(self) -> str:
        return "<ResultHandler: {}>".format(type(self).__name__)

    def write(self, result: Any) -> Any:
        return None

    def read(self, loc: str) -> Any:
        return None

    def __eq__(self, other: object) -> bool:
        """
        Equality depends on result handler type and any public attributes
        """
        if type(self) == type(other):
            assert isinstance(other, ResultHandler)  # mypy assert
            eq = True
            for attr in self.__dict__:
                if attr.startswith("_") or attr == "logger":
                    continue
                eq &= getattr(self, attr, object()) == getattr(other, attr, object())
            return eq
        return False



"""
A collection of tasks for interacting with Google Sheets.
"""
try:
    from prefect.tasks.gsheets.gsheets import (
        WriteGsheetRow,
        ReadGsheetRow,
    )
except ImportError:
    raise ImportError(
        'Using `prefect.tasks.gsheets` requires Prefect to be installed with the "gsheets" extra.'
    )



from typing import Any, Dict, List, Union

import azure.cosmos.cosmos_client

from prefect import Task
from prefect.client import Secret
from prefect.utilities.tasks import defaults_from_attrs


class CosmosDBCreateItem(Task):
    """
    Task for creating an item in a Azure Cosmos database.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    Args:
        - url (str, optional): The url to the database.
        - database_or_container_link (str, optional): link to the database or container.
        - item (dict, optional): the item to create
        - azure_credentials_secret (str, optional): the name of the Prefect Secret that stores
            your Azure credentials; this Secret must be JSON string with the key
            `AZ_COSMOS_AUTH`. The value should be dictionary containing `masterKey` or
            `resourceTokens`, where the `masterKey` value is the default authorization key to
            use to create the client, and `resourceTokens` value is the alternative
            authorization key.
        - options (dict, optional): options to be passed to the
            `azure.cosmos.cosmos_client.CosmosClient.CreateItem` method.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        url: str = None,
        database_or_container_link: str = None,
        item: Dict = None,
        azure_credentials_secret: str = "AZ_CREDENTIALS",
        options: Dict[Any, Any] = None,
        **kwargs
    ) -> None:
        self.url = url
        self.database_or_container_link = database_or_container_link
        self.item = item
        self.azure_credentials_secret = azure_credentials_secret
        self.options = options
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "url",
        "database_or_container_link",
        "item",
        "azure_credentials_secret",
        "options",
    )
    def run(
        self,
        url: str = None,
        database_or_container_link: str = None,
        item: Dict = None,
        azure_credentials_secret: str = "AZ_CREDENTIALS",
        options: Dict[Any, Any] = None,
    ) -> Dict[Any, Any]:
        """
        Task run method.

        Args:
            - url (str, optional): The url to the database.
            - database_or_container_link (str, optional): link to the database or container.
            - item (dict, optional): the item to create
            - azure_credentials_secret (str, optional): the name of the Prefect Secret
                that stores your Azure credentials; this Secret must be JSON string with the
                key `AZ_COSMOS_AUTH`. The value should be dictionary containing `masterKey` or
                `resourceTokens`, where the `masterKey` value is the default authorization key
                to use to create the client, and `resourceTokens` value is the alternative
                authorization key.
            - options (dict, optional): options to be passed to the
                `azure.cosmos.cosmos_client.CosmosClient.CreateItem` method.

        Returns:
            - (dict): the created item.
        """

        if url is None:
            raise ValueError("A url must be provided.")

        if database_or_container_link is None:
            raise ValueError("A database or container link must be provided.")

        if item is None:
            raise ValueError("An item must be provided.")

        azure_credentials = Secret(azure_credentials_secret).get()
        auth_dict = azure_credentials["AZ_COSMOS_AUTH"]

        client = azure.cosmos.cosmos_client.CosmosClient(
            url_connection=url, auth=auth_dict
        )

        return_item = client.CreateItem(
            database_or_container_link, item, options=options
        )

        return return_item


class CosmosDBReadItems(Task):
    """
    Task for reading items from a Azure Cosmos database.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    Args:
        - url (str, optional): The url to the database.
        - document_or_container_link (str, optional): link to a document or container.
            If a document link is provided, the document in question is returned, otherwise
            all docuements are returned.
        - azure_credentials_secret (str, optional): the name of the Prefect Secret
            that stores your Azure credentials; this Secret must be JSON string with the key
            `AZ_COSMOS_AUTH`. The value should be dictionary containing `masterKey` or
            `resourceTokens`, where the `masterKey` value is the default authorization key to
            use to create the client, and `resourceTokens` value is the alternative
            authorization key.
        - options (dict, optional): options to be passed to the
            `azure.cosmos.cosmos_client.CosmosClient.ReadItem` or `ReadItems` method.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        url: str = None,
        document_or_container_link: str = None,
        azure_credentials_secret: str = "AZ_CREDENTIALS",
        options: Dict[Any, Any] = None,
        **kwargs
    ) -> None:
        self.url = url
        self.document_or_container_link = document_or_container_link
        self.azure_credentials_secret = azure_credentials_secret
        self.options = options

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "url", "document_or_container_link", "azure_credentials_secret", "options"
    )
    def run(
        self,
        url: str = None,
        document_or_container_link: str = None,
        azure_credentials_secret: str = "AZ_CREDENTIALS",
        options: Dict[Any, Any] = None,
    ) -> Union[Dict[Any, Any], List[Dict[Any, Any]]]:
        """
        Task run method.

        Args:
            - url (str, optional): The url to the database.
            - document_or_container_link (str, optional): link to a document or container.
                If a document link is provided, the document in question is returned, otherwise
                all docuements are returned.
            - azure_credentials_secret (str, optional): the name of the Prefect Secret
                that stores your Azure credentials; this Secret must be JSON string with the
                key `AZ_COSMOS_AUTH`. The value should be dictionary containing `masterKey` or
                `resourceTokens`, where the `masterKey` value is the default authorization key
                to use to create the client, and `resourceTokens` value is the alternative
                authorization key.
            - options (dict, optional): options to be passed to the
                `azure.cosmos.cosmos_client.CosmosClient.ReadItem` or `ReadItems` method.

        Returns:
            - (dict or list)): a single document or all documents.
        """

        if url is None:
            raise ValueError("A url must be provided.")

        if document_or_container_link is None:
            raise ValueError("A document or container link must be provided.")

        azure_credentials = Secret(azure_credentials_secret).get()
        auth_dict = azure_credentials["AZ_COSMOS_AUTH"]

        client = azure.cosmos.cosmos_client.CosmosClient(
            url_connection=url, auth=auth_dict
        )

        if self._is_valid_document_link(document_or_container_link):
            return_obj = client.ReadItem(document_or_container_link, options=options)
        else:
            return_obj = client.ReadItems(
                document_or_container_link, feed_options=options
            )
            return_obj = list(return_obj)

        return return_obj

    @staticmethod
    def _is_valid_document_link(link: str) -> bool:
        trimmed_link = link.strip("/")
        split_link = trimmed_link.split("/")

        if (
            len(split_link) == 6
            and split_link[0] == "dbs"
            and split_link[2] == "colls"
            and split_link[4] == "docs"
        ):
            return True
        return False


class CosmosDBQueryItems(Task):
    """
    Task for creating an item in a Azure Cosmos database.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    Args:
        - url (str, optional): The url to the database.
        - database_or_container_link (str, optional): link to the database or container.
        - query (dict, optional): the query to run
        - azure_credentials_secret (str, optional): the name of the Prefect Secret
            that stores your Azure credentials; this Secret must be JSON string with the key
            `AZ_COSMOS_AUTH`. The value should be dictionary containing `masterKey` or
            `resourceTokens`, where the `masterKey` value is the default authorization key to
            use to create the client, and `resourceTokens` value is the alternative
            authorization key.
        - options (dict, optional): options to be passed to the
            `azure.cosmos.cosmos_client.CosmosClient.QueryItems` method.
        - partition_key (str, None): Partition key for the query.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        url: str = None,
        database_or_container_link: str = None,
        query: str = None,
        azure_credentials_secret: str = "AZ_CREDENTIALS",
        options: Dict[Any, Any] = None,
        partition_key: str = None,
        **kwargs
    ) -> None:
        self.url = url
        self.database_or_container_link = database_or_container_link
        self.query = query
        self.azure_credentials_secret = azure_credentials_secret
        self.options = options
        self.partition_key = partition_key
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "url",
        "database_or_container_link",
        "query",
        "azure_credentials_secret",
        "options",
        "partition_key",
    )
    def run(
        self,
        url: str = None,
        database_or_container_link: str = None,
        query: str = None,
        azure_credentials_secret: str = "AZ_CREDENTIALS",
        options: Dict[Any, Any] = None,
        partition_key: str = None,
    ) -> List:
        """
        Task run method.

        Args:
            - url (str, optional): The url to the database.
            - database_or_container_link (str, optional): link to the database or container.
            - query (dict, optional): the query to run
            - azure_credentials_secret (str, optional): the name of the Prefect Secret
                that stores your Azure credentials; this Secret must be JSON string with the
                key `AZ_COSMOS_AUTH`. The value should be dictionary containing `masterKey` or
                `resourceTokens`, where the `masterKey` value is the default authorization key
                to use to create the client, and `resourceTokens` value is the alternative
                authorization key.
            - options (dict, optional): options to be passed to the
                `azure.cosmos.cosmos_client.CosmosClient.QueryItems` method.
            - partition_key (str, None): Partition key for the query.

        Returns:
            - (list): a list containing the query results, one item per row.
        """

        if url is None:
            raise ValueError("A url must be provided.")

        if database_or_container_link is None:
            raise ValueError("A database or container link must be provided.")

        if query is None:
            raise ValueError("A query must be provided.")

        azure_credentials = Secret(azure_credentials_secret).get()
        auth_dict = azure_credentials["AZ_COSMOS_AUTH"]

        client = azure.cosmos.cosmos_client.CosmosClient(
            url_connection=url, auth=auth_dict
        )

        items = client.QueryItems(
            database_or_container_link,
            query,
            options=options,
            partition_key=partition_key,
        )

        return list(items)



import json

from prefect.tasks.secrets.base import SecretBase
from prefect.utilities.aws import get_boto_client
from prefect.utilities.tasks import defaults_from_attrs


class AWSSecretsManager(SecretBase):
    """
    Task for retrieving secrets from an AWS Secrets Manager and returning it as a dictionary.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    For authentication, there are two options: you can set the `AWS_CREDENTIALS` Prefect Secret
    containing your AWS access keys which will be passed directly to the `boto3` client, or you
    can [configure your flow's runtime
    environment](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#guide-configuration)
    for `boto3`.

    Args:
        - secret (str, optional): the name of the secret to retrieve
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(self, secret: str = None, **kwargs):
        self.secret = secret
        super().__init__(**kwargs)

    @defaults_from_attrs("secret")
    def run(self, secret: str = None, credentials: str = None) -> dict:
        """
        Task run method.

        Args:
            - secret (str): the name of the secret to retrieve
            - credentials (dict, optional): your AWS credentials passed from an upstream
                Secret task; this Secret must be a JSON string
                with two keys: `ACCESS_KEY` and `SECRET_ACCESS_KEY` which will be
                passed directly to `boto3`.  If not provided here or in context, `boto3`
                will fall back on standard AWS rules for authentication.

        Returns:
            - dict: the contents of this secret, as a dictionary
        """

        if secret is None:
            raise ValueError("A secret name must be provided.")

        secrets_client = get_boto_client("secretsmanager", credentials=credentials)

        secret_string = secrets_client.get_secret_value(SecretId=secret)["SecretString"]

        secret_dict = json.loads(secret_string)

        return secret_dict



import cloudpickle
import os
import string
import warnings

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional

from requests import Session
from requests.adapters import HTTPAdapter
from requests.models import Response
from requests.packages.urllib3.util.retry import Retry

from prefect.client import Secret
from prefect.environments.storage import Storage
from prefect.utilities.storage import extract_flow_from_file

if TYPE_CHECKING:
    from prefect.core.flow import Flow


class _SecretMapping(Mapping):
    """
    Magic mapping. When `__getitem__()` is called
    to look for the value of a `key`, this mapping will
    try to resolve it from the environment and, if not found
    there, from `prefect` secrets.

    Raises:
        - RuntimeError if the key you're looking for isn't found
    """

    def __getitem__(self, key: str) -> str:
        out = os.getenv(key, None)  # type: ignore
        if out is None:
            try:
                out = Secret(key).get()
            except ValueError:
                msg = (
                    "Template value '{}' does not refer to an "
                    "environment variable or Prefect secret."
                )
                raise RuntimeError(msg.format(key))
        return out  # type: ignore

    def __iter__(self) -> Iterator[Any]:
        return iter([])

    def __len__(self) -> int:
        return 0


_mapping = _SecretMapping()


def _render_dict(input_dict: Dict[str, Any]) -> Dict[str, Any]:
    """
    Replace string elements in a dictionary with environment variables
    or Prefect secrets.

    This method will look through all string values (not keys) of a dictionary,
    recursively, for template string. When it encounters a string like
    `${abc}`, it will first try to replace it with the value of environment
    variable `abc`. If environment variable `abc` is not defined, it will try
    to replace it with `prefect` secret `abc`. If that is also not found,
    raises a `KeyError`.

    Args:
        - input_dict (Dict[str, Any]): A dictionary that may contain
            template strings.

    Returns:
        - A dictionary with template strings replaced by the literal values of
            environment variables or Prefect secrets

    Raises:
        - RuntimeError if any template value cannot be found in the environment
            or `prefect` secrets
    """
    output_dict = {}

    for key, value in input_dict.items():
        if isinstance(value, str):
            new_value = string.Template(value).substitute(_mapping)
            output_dict[key] = new_value
        elif isinstance(value, dict):
            output_dict[key] = _render_dict(value)  # type: ignore
        else:
            output_dict[key] = value

    return output_dict


class Webhook(Storage):
    """
    Webhook storage class. This class represents the Storage interface for
    Flows stored and retrieved with HTTP requests.

    This storage class takes in keyword arguments which describe how to
    create the requests. These arguments' values can contain template
    strings which will be filled in dynamically from environment variables
    or Prefect secrets.

     **Note**: Flows registered with this Storage option will automatically be
     labeled with `webhook-flow-storage`.

    Args:
        - build_request_kwargs (dict): Dictionary of keyword arguments to the
            function from `requests` used to store the flow. Do not supply
            `"data"` to this argument, as it will be overwritten with the
            flow's content when `.build()` is run.
        - build_request_http_method (str): HTTP method identifying the type of
            request to execute when storing the flow. For example, `"POST"` for
            `requests.post()`.
        - get_flow_request_kwargs (dict): Dictionary of keyword arguments to
            the function from `requests` used to retrieve the flow.
        - get_flow_request_http_method (str): HTTP method identifying the type
            of request to execute when storing the flow. For example, `"GET"`
            for `requests.post()`.
        - stored_as_script (bool, optional): boolean for specifying if the
            flow has been stored as a `.py` file. Defaults to `False`.
        - flow_script_path (str, optional): path to a local `.py` file that
            defines the flow. You must pass a value to this argument if
            `stored_as_script` is `True`. This script's content will be read
            into a string and attached to the request in `build()` as UTF-8
            encoded binary data. Similarly, `.get_flow()` expects that the
            script's contents will be returned as binary data. This path will
            not be sent to Prefect Cloud and is only needed when running
            `.build()`.
        - **kwargs (Any, optional): any additional `Storage` initialization
            options

    Including Sensitive Data
    ------------------------

    It is common for requests used with this storage to need access to
    sensitive information.

    For example:

    - auth tokens passed in headers like `X-Api-Key` or `Authorization`
    - auth information passed in to URL as query parameters

    `Webhook` storage supports the inclusion of such sensitive information
    with templating. Any of the string values passed to
    `build_flow_request_kwargs` or `get_flow_request_kwargs` can include
    template strings like `${SOME_VARIABLE}`. When `.build()` or `.get_flow()`
    is run, such values will be replaced with the value of environment
    variables or, when no matching environment variable is found, Prefect
    Secrets.

    So, for example, to get an API key from an environment variable you
    can do the following

    ```python
    storage = Webhook(
        build_request_kwargs={
            "url": "some-service/upload",
            "headers" = {
                "Content-Type" = "application/octet-stream",
                "X-Api-Key": "${MY_COOL_ENV_VARIABLE}"
            }
        },
        build_request_http_method="POST",
    )
    ```

    You can also take advantage of this templating when only part
    of a string needs to be replaced.

    ```python
    storage = Webhook(
        get_flow_request_kwargs={
            "url": "some-service/download",
            "headers" = {
                "Accept" = "application/octet-stream",
                "Authorization": "Bearer ${MY_COOL_ENV_VARIABLE}"
            }
        },
        build_request_http_method="POST",
    )
    ```
    """

    def __init__(
        self,
        build_request_kwargs: Dict[str, Any],
        build_request_http_method: str,
        get_flow_request_kwargs: Dict[str, Any],
        get_flow_request_http_method: str,
        stored_as_script: bool = False,
        flow_script_path: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        self.flows = dict()  # type: Dict[str, str]
        self._flows = dict()  # type: Dict[str, "Flow"]

        # set up logic for authenticating with Saturn back-end service
        retry_logic = HTTPAdapter(max_retries=Retry(total=3))
        self._session = Session()
        self._session.mount("http://", retry_logic)
        self._session.mount("https://", retry_logic)

        self._method_to_function = {
            "GET": self._session.get,
            "PATCH": self._session.patch,
            "POST": self._session.post,
            "PUT": self._session.put,
        }

        if build_request_http_method not in self._method_to_function.keys():
            msg = "HTTP method '{}' not recognized".format(build_request_http_method)
            self.logger.fatal(msg)
            raise RuntimeError(msg)

        if get_flow_request_http_method not in self._method_to_function.keys():
            msg = "HTTP method '{}' not recognized".format(get_flow_request_http_method)
            self.logger.fatal(msg)
            raise RuntimeError(msg)

        self.stored_as_script = stored_as_script
        self.flow_script_path = flow_script_path

        self.build_request_kwargs = build_request_kwargs
        self.build_request_http_method = build_request_http_method

        self.get_flow_request_kwargs = get_flow_request_kwargs
        self.get_flow_request_http_method = get_flow_request_http_method

        self._build_responses: Optional[Dict[str, Response]] = None

        super().__init__(stored_as_script=stored_as_script, **kwargs)

    @property
    def default_labels(self) -> List[str]:
        return ["webhook-flow-storage"]

    def get_flow(self, flow_location: str = "placeholder") -> "Flow":
        """
        Get the flow from storage. This method will call
        `cloudpickle.loads()` on the binary content of the flow, so it
        should only be called in an environment with all of the flow's
        dependencies.

        Args:
            - flow_location (str): This argument is included to comply with the
                interface used by other storage objects, but it has no meaning
                for `Webhook` storage, since `Webhook` only corresponds to a
                single flow. Ignore it.

        Raises:
            - requests.exceptions.HTTPError if getting the flow fails
        """
        self.logger.info("Retrieving flow")
        req_function = self._method_to_function[self.get_flow_request_http_method]

        get_flow_request_kwargs = _render_dict(self.get_flow_request_kwargs)

        response = req_function(**get_flow_request_kwargs)  # type: ignore
        response.raise_for_status()

        if self.stored_as_script:
            flow_script_content = response.content.decode("utf-8")
            return extract_flow_from_file(file_contents=flow_script_content)  # type: ignore

        return cloudpickle.loads(response.content)

    def add_flow(self, flow: "Flow") -> str:
        """
        Method for adding a flow to a `Storage` object's in-memory
        storage. `.build()` will look here for flows.

        `Webhook` storage only supports a single flow per storage
        object, so this method will overwrite any existing flows
        stored in an instance.

        Args:
            - flow (Flow): a Prefect Flow to add

        Returns:
            - str: the name of the flow
        """
        self.flows = {flow.name: flow.name}
        self._flows = {flow.name: flow}
        return flow.name

    def build(self) -> "Webhook":
        """
        Build the Webhook storage object by issuing an HTTP request
        to store the flow.

        If `self.stored_as_script` is `True`, this method
        will read in the contents of `self.flow_script_path`, convert it to
        byes, and attach it to the request as `data`.

        The response from this request is stored in `._build_responses`,
        a dictionary keyed by flow name. If you are using a service where
        all the details necessary to fetch a flow cannot be known until you've
        stored it, you can do something like the following.

        ```python
        import cloudpickle
        import json
        import os
        import random
        import requests

        from prefect import task, Task, Flow
        from prefect.environments.storage import Webhook

        @task
        def random_number():
            return random.randint(0, 100)
        with Flow("test-flow") as flow:
            random_number()


        flow.storage = Webhook(
            build_request_kwargs={
                "url": "some-service/upload",
                "headers": {"Content-Type": "application/octet-stream"},
            },
            build_request_http_method="POST",
            get_flow_request_kwargs={
                "url": "some-service/download",
                "headers": {"Accept": "application/octet-stream"},
            },
            get_flow_request_http_method="GET",
        )

        flow.storage.add_flow(flow)
        res = flow.storage.build()

        # get the ID from the response
        flow_id = res._build_responses[flow.name].json()["id"]

        #  update storage
        flow.storage.get_flow_request_kwargs["url"] = f"{GET_ROUTE}/{flow_id}"
        ```

        Returns:
            - Storage: a Webhook storage object

        Raises:
            - requests.exceptions.HTTPError if pushing the flow fails
        """
        self.run_basic_healthchecks()
        self._build_responses = {}

        for flow_name, flow in self._flows.items():
            self.logger.info("Uploading flow '{}'".format(flow_name))

            data = cloudpickle.dumps(flow)
            if self.stored_as_script:

                # these checks are here in build() instead of the constructor
                # so that serialization and deserialization of flows doesnot fail
                if not self.flow_script_path:
                    msg = "flow_script_path must be provided if stored_as_script=True"
                    self.logger.fatal(msg)
                    raise RuntimeError(msg)

                if not os.path.isfile(self.flow_script_path):
                    msg = "file '{}' passed to flow_script_path does not exist".format(
                        self.flow_script_path
                    )
                    self.logger.fatal(msg)
                    raise RuntimeError(msg)

                with open(self.flow_script_path, "r") as f:
                    data = f.read().encode("utf-8")

            req_function = self._method_to_function[self.build_request_http_method]

            build_request_kwargs = _render_dict(self.build_request_kwargs)

            if "data" in build_request_kwargs.keys():
                msg = (
                    "'data' found in build_request_kwargs. This value is "
                    "overwritten with the flow content and should not "
                    "be set directly"
                )
                self.logger.warning(msg)
                warnings.warn(msg, RuntimeWarning)
            build_request_kwargs["data"] = data

            response = req_function(**build_request_kwargs)  # type: ignore
            response.raise_for_status()

            self._build_responses[flow_name] = response
            self.logger.info("Successfully uploaded flow '{}'".format(flow_name))

        return self

    def __contains__(self, obj: Any) -> bool:
        """
        Method for determining whether an object is
        contained within this storage.
        """
        if not isinstance(obj, str):
            return False
        return obj in self.flows



from prefect import Task
from prefect.client import Secret
from prefect.utilities.tasks import defaults_from_attrs
from typing import cast
from typing import Any

try:
    from jira import JIRA
except ImportError:
    pass


class JiraTask(Task):
    """
    Task for creating a jira issue. For this task to function properly, you need a Jira account
    and API token.  The API token can be created at: https://id.atlassian.com/manage/api-tokens
    The Jira account username ('JIRAUSER'), API token ('JIRATOKEN') can be set as part of a
    'JIRASECRETS' object in Prefect Secrets.

    An example 'JIRASECRETS' secret configuration looks like:

    ```toml
    [secrets]
    JIRASECRETS.JIRATOKEN = "XXXXXXXXX"
    JIRASECRETS.JIRAUSER = "xxxxx@yyy.com"
    JIRASECRETS.JIRASERVER = "https://???.atlassian.net"
    ```

    The server URL can be set as part of the 'JIRASECRETS' object ('JIRASERVER') or passed to
    the task as the "server_URL" argument.

    Args:
        - server_url (str): the URL of your atlassian account e.g.
            "https://test.atlassian.net".  Can also be set as a Prefect Secret.
        - project_name(str):  the key for your jira project. Can also be set at run time.
        - assignee (str, optional): the atlassian accountId of the person you want to assign
            the ticket to.  Defaults to "automatic" if this is not set. Can also be set at run
            time.
        - issue_type (str, optional): the type of issue you want to create.  Can also be set at
            run time. Defaults to 'Task'.
        - summary (str, optional): summary or title for your issue. Can also be set at run time.
        - description (str, optional): description or additional information for the issue. Can
            also be set at run time.
        - **kwargs (Any, optional): additional keyword arguments to pass to the standard Task
            init method.
    """

    def __init__(
        self,
        server_url: str = None,
        project_name: str = None,
        assignee: str = "-1",
        issue_type: str = None,
        summary: str = None,
        description: str = None,
        **kwargs: Any
    ):
        self.server_url = server_url
        self.project_name = project_name
        self.assignee = assignee
        self.issue_type = issue_type
        self.summary = summary
        self.description = description
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "server_url", "project_name", "assignee", "issue_type", "summary", "description"
    )
    def run(
        self,
        username: str = None,
        access_token: str = None,
        server_url: str = None,
        project_name: str = None,
        assignee: str = "-1",
        issue_type: str = None,
        summary: str = None,
        description: str = None,
    ) -> None:
        """
        Run method for this Task. Invoked by calling this Task after initialization within a
        Flow context, or by using `Task.bind`.

        Args:
            - username(str): the jira username, provided with a Prefect secret (defaults to
                JIRAUSER in JIRASECRETS)
            - access_token (str): a Jira access token, provided with a Prefect secret (defaults
                to JIRATOKEN in JIRASECRETS)
            - server_url (str): the URL of your atlassian account e.g.
                "https://test.atlassian.net".  Can also be set as a Prefect Secret. Defaults to
                the one provided at initialization
            - project_name(str):  the key for your jira project; defaults to the one provided
                at initialization
            - assignee (str, optional): the atlassian accountId of the person you want to
                assign the ticket to; defaults to "automatic" if this is not set; defaults to
                the one provided at initialization
            - issue_type (str, optional): the type of issue you want to create; defaults to
                'Task'
            - summary (str, optional): summary or title for your issue; defaults to the one
                provided at initialization
            - description (str, optional): description or additional information for the issue;
                defaults to the one provided at initialization

        Raises:
            - ValueError: if a `project_name` or 'summary' are not provided

        Returns:
            - None
        """

        jira_credentials = cast(dict, Secret("JIRASECRETS").get())

        if username is None:
            username = jira_credentials["JIRAUSER"]

        if access_token is None:
            access_token = jira_credentials["JIRATOKEN"]

        if server_url is None:
            server_url = jira_credentials["JIRASERVER"]

        if issue_type is None:
            issue_type = "Task"

        if project_name is None:
            raise ValueError("A project name must be provided")

        if summary is None:
            raise ValueError("A summary must be provided")

        jira = JIRA(basic_auth=(username, access_token), options={"server": server_url})

        options = {
            "project": project_name,
            "assignee": {"accountId": assignee},
            "issuetype": {"name": issue_type},
            "summary": summary,
            "description": description,
        }
        created = jira.create_issue(options)

        if not created:
            raise ValueError("Creating Jira Issue failed")



from typing import Any

from marshmallow import fields, post_load

from prefect.environments.storage import (
    GCS,
    S3,
    Azure,
    Docker,
    Local,
    Storage,
    GitHub,
    Webhook,
)
from prefect.utilities.serialization import JSONCompatible, ObjectSchema, OneOfSchema


class AzureSchema(ObjectSchema):
    class Meta:
        object_class = Azure

    container = fields.String(allow_none=False)
    blob_name = fields.String(allow_none=True)
    stored_as_script = fields.Bool(allow_none=True)
    flows = fields.Dict(key=fields.Str(), values=fields.Str())
    secrets = fields.List(fields.Str(), allow_none=True)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> Azure:
        flows = data.pop("flows", dict())
        base_obj = super().create_object(data)
        base_obj.flows = flows
        return base_obj


class BaseStorageSchema(ObjectSchema):
    class Meta:
        object_class = Storage


class DockerSchema(ObjectSchema):
    class Meta:
        object_class = Docker

    registry_url = fields.String(allow_none=True)
    image_name = fields.String(allow_none=True)
    image_tag = fields.String(allow_none=True)
    path = fields.Str(allow_none=True)
    stored_as_script = fields.Bool(allow_none=True)
    flows = fields.Dict(key=fields.Str(), values=fields.Str())
    prefect_version = fields.String(allow_none=False)
    secrets = fields.List(fields.Str(), allow_none=True)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> Docker:
        flows = data.pop("flows", dict())
        base_obj = super().create_object(data)
        base_obj.flows = flows
        return base_obj


class GCSSchema(ObjectSchema):
    class Meta:
        object_class = GCS

    bucket = fields.Str(allow_none=False)
    key = fields.Str(allow_none=True)
    project = fields.Str(allow_none=True)
    stored_as_script = fields.Bool(allow_none=True)
    flows = fields.Dict(key=fields.Str(), values=fields.Str())
    secrets = fields.List(fields.Str(), allow_none=True)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> GCS:
        flows = data.pop("flows", dict())
        base_obj = super().create_object(data)
        base_obj.flows = flows
        return base_obj


class LocalSchema(ObjectSchema):
    class Meta:
        object_class = Local

    directory = fields.Str(allow_none=False)
    path = fields.Str(allow_none=True)
    stored_as_script = fields.Bool(allow_none=True)
    flows = fields.Dict(key=fields.Str(), values=fields.Str())
    secrets = fields.List(fields.Str(), allow_none=True)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> Docker:
        flows = data.pop("flows", dict())
        data.update(validate=False)
        base_obj = super().create_object(data)
        base_obj.flows = flows
        return base_obj


class S3Schema(ObjectSchema):
    class Meta:
        object_class = S3

    bucket = fields.String(allow_none=False)
    key = fields.String(allow_none=True)
    stored_as_script = fields.Bool(allow_none=True)
    flows = fields.Dict(key=fields.Str(), values=fields.Str())
    client_options = fields.Dict(
        key=fields.Str(), values=JSONCompatible(), allow_none=True
    )
    secrets = fields.List(fields.Str(), allow_none=True)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> S3:
        flows = data.pop("flows", dict())
        base_obj = super().create_object(data)
        base_obj.flows = flows
        return base_obj


class GitHubSchema(ObjectSchema):
    class Meta:
        object_class = GitHub

    repo = fields.String(allow_none=False)
    path = fields.String(allow_none=True)
    flows = fields.Dict(key=fields.Str(), values=fields.Str())
    secrets = fields.List(fields.Str(), allow_none=True)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> GitHub:
        flows = data.pop("flows", dict())
        base_obj = super().create_object(data)
        base_obj.flows = flows
        return base_obj


class WebhookSchema(ObjectSchema):
    class Meta:
        object_class = Webhook

    build_request_kwargs = fields.Dict(key=fields.Str, allow_none=False)
    build_request_http_method = fields.String(allow_none=False)
    get_flow_request_kwargs = fields.Dict(key=fields.Str, allow_none=False)
    get_flow_request_http_method = fields.String(allow_none=False)
    stored_as_script = fields.Bool(allow_none=True)
    flows = fields.Dict(key=fields.Str(), values=fields.Str())
    secrets = fields.List(fields.Str(), allow_none=True)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> Webhook:
        flows = data.pop("flows", dict())
        base_obj = super().create_object(data)
        base_obj.flows = flows
        return base_obj


class StorageSchema(OneOfSchema):
    """
    Field that chooses between several nested schemas
    """

    # map class name to schema
    type_schemas = {
        "Azure": AzureSchema,
        "Docker": DockerSchema,
        "GCS": GCSSchema,
        "Local": LocalSchema,
        "Storage": BaseStorageSchema,
        "S3": S3Schema,
        "GitHub": GitHubSchema,
        "Webhook": WebhookSchema,
    }



"""
Secret Tasks are a special kind of Prefect Task used to represent the retrieval of sensitive data.

The base implementation uses Prefect Cloud secrets, but users are encouraged to subclass the `Secret` task
class for interacting with other secret providers. Secrets always use a special kind of result handler that
prevents the persistence of sensitive information.
"""
from .base import SecretBase, PrefectSecret, Secret
from .env_var import EnvVarSecret



"""
The tasks in this module can be used to represent arbitrary functions.

In general, users will not instantiate these tasks by hand; they will
automatically be applied when users apply the `@task` decorator.
"""

from typing import Any, Callable

import prefect


class _DocProxy(object):
    """A descriptor that proxies through the docstring for the wrapped task as
    the docstring for a `FunctionTask` instance."""

    def __init__(self, cls_doc):
        self._cls_doc = cls_doc

    def __get__(self, obj, cls):
        if obj is None:
            return self._cls_doc
        else:
            return getattr(obj.run, "__doc__", None) or self._cls_doc


class FunctionTask(prefect.Task):
    __doc__ = _DocProxy(
        """A convenience Task for functionally creating Task instances with
    arbitrary callable `run` methods.

    Args:
        - fn (callable): the function to be the task's `run` method
        - name (str, optional): the name of this task
        - **kwargs: keyword arguments that will be passed to the Task
            constructor

    Raises:
        - ValueError: if the provided function violates signature requirements
            for Task run methods

    Example:
    ```python
    task = FunctionTask(lambda x: x - 42, name="Subtract 42")

    with Flow("My Flow") as f:
        result = task(42)
    ```
    """
    )

    def __init__(self, fn: Callable, name: str = None, **kwargs: Any):
        if not callable(fn):
            raise TypeError("fn must be callable.")

        # set the name from the fn
        if name is None:
            name = getattr(fn, "__name__", type(self).__name__)

        prefect.core.task._validate_run_signature(fn)  # type: ignore
        self.run = fn

        super().__init__(name=name, **kwargs)

    def __getattr__(self, k):
        if k == "__wrapped__":
            return self.run
        raise AttributeError(f"'FunctionTask' object has no attribute {k}")



"""
This module contains a collection of tasks for interacting with AWS resources.

All AWS related tasks can be authenticated using the `AWS_CREDENTIALS` Prefect Secret that should be a dictionary with two keys: `"ACCESS_KEY"` and `"SECRET_ACCESS_KEY"`.  See [Third Party Authentication](../../../orchestration/recipes/third_party_auth.html) for more information.
"""
try:
    from prefect.tasks.aws.s3 import S3Download, S3Upload
    from prefect.tasks.aws.lambda_function import (
        LambdaCreate,
        LambdaDelete,
        LambdaInvoke,
        LambdaList,
    )
    from prefect.tasks.aws.step_function import StepActivate
    from prefect.tasks.aws.secrets_manager import AWSSecretsManager
except ImportError:
    raise ImportError(
        'Using `prefect.tasks.aws` requires Prefect to be installed with the "aws" extra.'
    )



import multiprocessing
import ntpath
import posixpath
import re
import sys
from sys import platform
from typing import TYPE_CHECKING, Dict, Iterable, List, Tuple

from prefect import config, context
from prefect.agent import Agent
from prefect.utilities.agent import get_flow_image
from prefect.utilities.docker_util import get_docker_ip
from prefect.utilities.graphql import GraphQLResult

if TYPE_CHECKING:
    import docker


class DockerAgent(Agent):
    """
    Agent which deploys flow runs locally as Docker containers. Information on using the
    Docker Agent can be found at https://docs.prefect.io/orchestration/agents/docker.html

    Environment variables may be set on the agent to be provided to each flow run's container:
    ```
    prefect agent start docker --env MY_SECRET_KEY=secret --env OTHER_VAR=$OTHER_VAR
    ```

    The default Docker daemon may be overridden by providing a different `base_url`:
    ```
    prefect agent start docker --base-url "tcp://0.0.0.0:2375"
    ```

    Args:
        - name (str, optional): An optional name to give this agent. Can also be set through
            the environment variable `PREFECT__CLOUD__AGENT__NAME`. Defaults to "agent"
        - labels (List[str], optional): a list of labels, which are arbitrary string
            identifiers used by Prefect Agents when polling for work
        - env_vars (dict, optional): a dictionary of environment variables and values that will
            be set on each flow run that this agent submits for execution
        - max_polls (int, optional): maximum number of times the agent will poll Prefect Cloud
            for flow runs; defaults to infinite
        - agent_address (str, optional):  Address to serve internal api at. Currently this is
            just health checks for use by an orchestration layer. Leave blank for no api server
            (default).
        - no_cloud_logs (bool, optional): Disable logging to a Prefect backend for this agent
            and all deployed flow runs
        - base_url (str, optional): URL for a Docker daemon server. Defaults to
            `unix:///var/run/docker.sock` however other hosts such as
            `tcp://0.0.0.0:2375` can be provided
        - no_pull (bool, optional): Flag on whether or not to pull flow images.
            Defaults to `False` if not provided here or in context.
        - show_flow_logs (bool, optional): a boolean specifying whether the agent should
            re-route Flow run logs to stdout; defaults to `False`
        - volumes (List[str], optional): a list of Docker volume mounts to be attached to any
            and all created containers.
        - network (str, optional): Add containers to an existing docker network
        - docker_interface (bool, optional): Toggle whether or not a `docker0` interface is
            present on this machine.  Defaults to `True`. **Note**: This is mostly relevant for
            some Docker-in-Docker setups that users may be running their agent with.
        - reg_allow_list (List[str], optional): Limits Docker Agent to only pull images
            from the listed registries.
    """

    def __init__(
        self,
        name: str = None,
        labels: Iterable[str] = None,
        env_vars: dict = None,
        max_polls: int = None,
        agent_address: str = None,
        no_cloud_logs: bool = False,
        base_url: str = None,
        no_pull: bool = None,
        volumes: List[str] = None,
        show_flow_logs: bool = False,
        network: str = None,
        docker_interface: bool = True,
        reg_allow_list: List[str] = None,
    ) -> None:
        super().__init__(
            name=name,
            labels=labels,
            env_vars=env_vars,
            max_polls=max_polls,
            agent_address=agent_address,
            no_cloud_logs=no_cloud_logs,
        )
        if platform == "win32":
            default_url = "npipe:////./pipe/docker_engine"
        else:
            default_url = "unix://var/run/docker.sock"
        self.logger.debug(
            "Platform {} and default docker daemon {}".format(platform, default_url)
        )

        # Determine Daemon URL
        self.base_url = base_url or context.get("base_url", default_url)
        self.logger.debug("Base docker daemon url {}".format(self.base_url))

        # Determine pull specification
        self.no_pull = no_pull or context.get("no_pull", False)
        self.logger.debug("no_pull set to {}".format(self.no_pull))

        # Resolve volumes from specs
        (
            self.named_volumes,
            self.container_mount_paths,
            self.host_spec,
        ) = self._parse_volume_spec(volumes or [])

        # Add containers to a docker network
        self.network = network
        self.logger.debug("Docker network set to {}".format(self.network))

        self.docker_interface = docker_interface
        self.logger.debug(
            "Docker interface toggle set to {}".format(self.docker_interface)
        )

        self.failed_connections = 0
        self.docker_client = self._get_docker_client()
        self.show_flow_logs = show_flow_logs
        self.processes = []  # type: List[multiprocessing.Process]

        self.reg_allow_list = reg_allow_list
        self.logger.debug("reg_allow_list set to {}".format(self.reg_allow_list))

        # Ping Docker daemon for connection issues
        try:
            self.logger.debug("Pinging docker daemon")
            self.docker_client.ping()
        except Exception as exc:
            self.logger.exception(
                "Issue connecting to the Docker daemon. Make sure it is running."
            )
            raise exc

        self.logger.debug(f"Base URL: {self.base_url}")
        self.logger.debug(f"No pull: {self.no_pull}")
        self.logger.debug(f"Volumes: {volumes}")
        self.logger.debug(f"Network: {self.network}")
        self.logger.debug(f"Docker interface: {self.docker_interface}")

    def _get_docker_client(self) -> "docker.APIClient":
        # 'import docker' is expensive time-wise, we should do this just-in-time to keep
        # the 'import prefect' time low
        import docker

        return docker.APIClient(base_url=self.base_url, version="auto")

    def heartbeat(self) -> None:
        try:
            if not self.docker_client.ping():
                raise RuntimeError("Unexpected Docker ping result")
            if self.failed_connections > 0:
                self.logger.info("Reconnected to Docker daemon")
            self.failed_connections = 0
        except Exception as exc:
            self.logger.warning("Failed heartbeat: {}".format(repr(exc)))
            self.failed_connections += 1

        if self.failed_connections >= 6:
            self.logger.error(
                "Cannot reconnect to Docker daemon. Agent is shutting down."
            )
            raise SystemExit()

    def on_shutdown(self) -> None:
        """
        Cleanup any child processes created for streaming logs. This is to prevent
        logs from displaying on the terminal after the agent exits.
        """
        for proc in self.processes:
            if proc.is_alive():
                proc.terminate()

    def _is_named_volume_unix(self, canditate_path: str) -> bool:
        if not canditate_path:
            return False

        return not canditate_path.startswith((".", "/", "~"))

    def _is_named_volume_win32(self, canditate_path: str) -> bool:
        result = self._is_named_volume_unix(canditate_path)

        return (
            result
            and not re.match(r"^[A-Za-z]\:\\.*", canditate_path)
            and not canditate_path.startswith("\\")
        )

    def _parse_volume_spec(
        self, volume_specs: List[str]
    ) -> Tuple[Iterable[str], Iterable[str], Dict[str, Dict[str, str]]]:
        if platform == "win32":
            return self._parse_volume_spec_win32(volume_specs)
        return self._parse_volume_spec_unix(volume_specs)

    def _parse_volume_spec_win32(
        self, volume_specs: List[str]
    ) -> Tuple[Iterable[str], Iterable[str], Dict[str, Dict[str, str]]]:
        named_volumes = []  # type: List[str]
        container_mount_paths = []  # type: List[str]
        host_spec = {}  # type: Dict[str, Dict[str, str]]

        for volume_spec in volume_specs:
            fields = volume_spec.split(":")

            if fields[-1] in ("ro", "rw"):
                mode = fields.pop()
            else:
                mode = "rw"

            if len(fields) == 3 and len(fields[0]) == 1:
                # C:\path1:/path2   <-- extenal and internal path
                external = ntpath.normpath(":".join(fields[0:2]))
                internal = posixpath.normpath(fields[2])
            elif len(fields) == 2:
                combined_path = ":".join(fields)
                (drive, path) = ntpath.splitdrive(combined_path)
                if drive:
                    # C:\path1          <-- assumed container path of /path1
                    external = ntpath.normpath(combined_path)

                    # C:\path1  --> /c/path1
                    path = str("/" + drive.lower().rstrip(":") + path).replace(
                        "\\", "/"
                    )
                    internal = posixpath.normpath(path)
                else:
                    # /path1:\path2     <-- extenal and internal path (relative to current drive)
                    # C:/path2          <-- valid named volume
                    external = ntpath.normpath(fields[0])
                    internal = posixpath.normpath(fields[1])
            elif len(fields) == 1:
                # \path1          <-- assumed container path of /path1 (relative to current drive)
                external = ntpath.normpath(fields[0])
                internal = external
            else:
                raise ValueError(
                    "Unable to parse volume specification '{}'".format(volume_spec)
                )

            container_mount_paths.append(internal)

            if external and self._is_named_volume_win32(external):
                named_volumes.append(external)
                if mode != "rw":
                    raise ValueError(
                        "Named volumes can only have 'rw' mode, provided '{}'".format(
                            mode
                        )
                    )
            else:
                if not external:
                    # no internal container path given, assume the host path is the same as the
                    # internal path
                    external = internal
                host_spec[external] = {
                    "bind": internal,
                    "mode": mode,
                }

        return named_volumes, container_mount_paths, host_spec

    def _parse_volume_spec_unix(
        self, volume_specs: List[str]
    ) -> Tuple[Iterable[str], Iterable[str], Dict[str, Dict[str, str]]]:
        named_volumes = []  # type: List[str]
        container_mount_paths = []  # type: List[str]
        host_spec = {}  # type: Dict[str, Dict[str, str]]

        for volume_spec in volume_specs:
            fields = volume_spec.split(":")

            if len(fields) > 3:
                raise ValueError(
                    f"Docker volume format is invalid: {volume_spec} "
                    f"(should be 'external:internal[:mode]')"
                )

            if len(fields) == 1:
                external = None
                internal = posixpath.normpath(fields[0].strip())
            else:
                external = posixpath.normpath(fields[0].strip())
                internal = posixpath.normpath(fields[1].strip())

            mode = "rw"
            if len(fields) == 3:
                mode = fields[2]

            container_mount_paths.append(internal)

            if external and self._is_named_volume_unix(external):
                named_volumes.append(external)
                if mode != "rw":
                    raise ValueError(
                        "Named volumes can only have 'rw' mode, provided '{}'".format(
                            mode
                        )
                    )
            else:
                if not external:
                    # no internal container path given, assume the host path is the same as the
                    # internal path
                    external = internal
                host_spec[external] = {
                    "bind": internal,
                    "mode": mode,
                }

        return named_volumes, container_mount_paths, host_spec

    def deploy_flow(self, flow_run: GraphQLResult) -> str:
        """
        Deploy flow runs on your local machine as Docker containers

        Args:
            - flow_run (GraphQLResult): A GraphQLResult flow run object

        Returns:
            - str: Information about the deployment
        """
        self.logger.info(
            "Deploying flow run {}".format(flow_run.id)  # type: ignore
        )

        # 'import docker' is expensive time-wise, we should do this just-in-time to keep
        # the 'import prefect' time low
        import docker

        image = get_flow_image(flow_run=flow_run)
        env_vars = self.populate_env_vars(flow_run=flow_run)

        if not self.no_pull and len(image.split("/")) > 1:
            self.logger.info("Pulling image {}...".format(image))
            registry = image.split("/")[0]
            if self.reg_allow_list and registry not in self.reg_allow_list:
                self.logger.error(
                    "Trying to pull image from a Docker registry '{}' which"
                    " is not in the reg_allow_list".format(registry)
                )
                raise ValueError(
                    "Trying to pull image from a Docker registry '{}' which"
                    " is not in the reg_allow_list".format(registry)
                )
            else:
                pull_output = self.docker_client.pull(image, stream=True, decode=True)
                for line in pull_output:
                    self.logger.debug(line)
                self.logger.info("Successfully pulled image {}...".format(image))

        # Create any named volumes (if they do not already exist)
        for named_volume_name in self.named_volumes:
            try:
                self.docker_client.inspect_volume(name=named_volume_name)
            except docker.errors.APIError:
                self.logger.debug("Creating named volume {}".format(named_volume_name))
                self.docker_client.create_volume(
                    name=named_volume_name,
                    driver="local",
                    labels={"prefect_created": "true"},
                )

        # Create a container
        self.logger.debug("Creating Docker container {}".format(image))

        host_config = {"auto_remove": True}  # type: dict
        container_mount_paths = self.container_mount_paths
        if container_mount_paths:
            host_config.update(binds=self.host_spec)

        if sys.platform.startswith("linux") and self.docker_interface:
            docker_internal_ip = get_docker_ip()
            host_config.update(extra_hosts={"host.docker.internal": docker_internal_ip})

        networking_config = None
        if self.network:
            networking_config = self.docker_client.create_networking_config(
                {self.network: self.docker_client.create_endpoint_config()}
            )

        container = self.docker_client.create_container(
            image,
            command="prefect execute cloud-flow",
            environment=env_vars,
            volumes=container_mount_paths,
            host_config=self.docker_client.create_host_config(**host_config),
            networking_config=networking_config,
        )

        # Start the container
        self.logger.debug(
            "Starting Docker container with ID {}".format(container.get("Id"))
        )
        if self.network:
            self.logger.debug(
                "Adding container to docker network: {}".format(self.network)
            )

        self.docker_client.start(container=container.get("Id"))

        if self.show_flow_logs:
            proc = multiprocessing.Process(
                target=self.stream_container_logs,
                kwargs={"container_id": container.get("Id")},
            )

            proc.start()
            self.processes.append(proc)

        self.logger.debug("Docker container {} started".format(container.get("Id")))

        return "Container ID: {}".format(container.get("Id"))

    def stream_container_logs(self, container_id: str) -> None:
        """
        Stream container logs back to stdout

        Args:
            - container_id (str): ID of a container to stream logs
        """
        for log in self.docker_client.logs(
            container=container_id, stream=True, follow=True
        ):
            print(str(log, "utf-8").rstrip())

    def populate_env_vars(self, flow_run: GraphQLResult) -> dict:
        """
        Populate metadata and variables in the environment variables for a flow run

        Args:
            - flow_run (GraphQLResult): A flow run object

        Returns:
            - dict: a dictionary representing the populated environment variables
        """
        if "localhost" in config.cloud.api:
            api = "http://host.docker.internal:{}".format(config.server.port)
        else:
            api = config.cloud.api

        return {
            "PREFECT__CLOUD__API": api,
            "PREFECT__CLOUD__AUTH_TOKEN": config.cloud.agent.auth_token,
            "PREFECT__CLOUD__AGENT__LABELS": str(self.labels),
            "PREFECT__CONTEXT__FLOW_RUN_ID": flow_run.id,  # type: ignore
            "PREFECT__CONTEXT__FLOW_ID": flow_run.flow.id,  # type: ignore
            "PREFECT__CLOUD__USE_LOCAL_SECRETS": "false",
            "PREFECT__LOGGING__LOG_TO_CLOUD": str(self.log_to_cloud).lower(),
            "PREFECT__LOGGING__LEVEL": "DEBUG",
            "PREFECT__ENGINE__FLOW_RUNNER__DEFAULT_CLASS": "prefect.engine.cloud.CloudFlowRunner",
            "PREFECT__ENGINE__TASK_RUNNER__DEFAULT_CLASS": "prefect.engine.cloud.CloudTaskRunner",
            **self.env_vars,
        }


if __name__ == "__main__":
    DockerAgent().start()



from typing import Any

import airtable

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class WriteAirtableRow(Task):
    """
    A task for writing a row to an Airtable table.

    Note that _all_ initialization settings can be provided / overwritten at runtime.

    Args:
        - base_key (str): the Airtable base key
        - table_name (str): the table name
        - **kwargs (optional): additional kwargs to pass to the `Task` constructor
    """

    def __init__(self, base_key: str = None, table_name: str = None, **kwargs: Any):
        self.base_key = base_key
        self.table_name = table_name
        super().__init__(**kwargs)

    @defaults_from_attrs("base_key", "table_name")
    def run(
        self,
        data: dict,
        base_key: str = None,
        table_name: str = None,
        api_key: str = None,
    ) -> dict:
        """
        Inserts data into an Airtable table

        Args:
            - data (dict): the data to insert. This should be formatted as a dictionary mapping
                each column name to a value.
            - base_key (str): the Airtable base key
            - table_name (str): the table name
            - api_key (str): an Airtable API key. This can be provided via a Prefect Secret

        Returns:
            - a dictionary containing information about the successful insert
        """
        table = airtable.Airtable(base_key, table_name, api_key)
        return table.insert(data)


class ReadAirtableRow(Task):
    """
    A task for reading a row from an Airtable table.

    Note that _all_ initialization settings can be provided / overwritten at runtime.

    Args:
        - base_key (str): the Airtable base key
        - table_name (str): the table name
        - **kwargs (optional): additional kwargs to pass to the `Task` constructor
    """

    def __init__(self, base_key: str = None, table_name: str = None, **kwargs: Any):
        self.base_key = base_key
        self.table_name = table_name
        super().__init__(**kwargs)

    @defaults_from_attrs("base_key", "table_name")
    def run(
        self,
        id: str,
        base_key: str = None,
        table_name: str = None,
        api_key: str = None,
    ) -> dict:
        """
        Reads a row an Airtable table by its id

        Args:
            - id (str): the id of the row
            - base_key (str): the Airtable base key
            - table_name (str): the table name
            - api_key (str): an Airtable API key. This can be provided via a Prefect Secret

        Returns:
            - a dictionary with the keys as the columns and the values as the row's values
        """
        table = airtable.Airtable(base_key, table_name, api_key)
        return table.get(id)



from typing import Any, Set

from marshmallow import fields, post_load, utils

import prefect
from prefect.serialization.edge import EdgeSchema
from prefect.serialization.environment import EnvironmentSchema
from prefect.serialization.schedule import ScheduleSchema
from prefect.serialization.storage import StorageSchema
from prefect.serialization.task import ParameterSchema, TaskSchema
from prefect.utilities.serialization import (
    Nested,
    ObjectSchema,
    to_qualified_name,
)


def get_parameters(obj: prefect.Flow, context: dict) -> Set:
    if isinstance(obj, prefect.Flow):
        return {p for p in obj.tasks if isinstance(p, prefect.core.parameter.Parameter)}
    else:
        return utils.get_value(obj, "parameters")


def get_reference_tasks(obj: prefect.Flow, context: dict) -> Set:
    if isinstance(obj, prefect.Flow):
        return obj._reference_tasks
    else:
        return utils.get_value(obj, "reference_tasks")


class FlowSchema(ObjectSchema):
    class Meta:
        object_class = lambda: prefect.core.Flow
        exclude_fields = ["type", "parameters"]
        # ordered to make sure Task objects are loaded before Edge objects, due to Task caching
        ordered = True

    project = fields.String(allow_none=True)
    name = fields.String(required=True, allow_none=True)
    version = fields.String(allow_none=True)
    description = fields.String(allow_none=True)
    type = fields.Function(lambda flow: to_qualified_name(type(flow)), lambda x: x)
    schedule = fields.Nested(ScheduleSchema, allow_none=True)
    parameters = Nested(ParameterSchema, value_selection_fn=get_parameters, many=True)
    tasks = fields.Nested(TaskSchema, many=True)
    edges = fields.Nested(EdgeSchema, many=True)
    reference_tasks = Nested(
        TaskSchema, value_selection_fn=get_reference_tasks, many=True, only=["slug"]
    )
    environment = fields.Nested(EnvironmentSchema, allow_none=True)
    storage = fields.Nested(StorageSchema, allow_none=True)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> prefect.core.Flow:
        """
        Flow edges are validated, for example to make sure the keys match Task inputs,
        but because we are deserializing all Tasks as base Tasks, the edge validation will
        fail (base Tasks have no inputs). Therefore we hold back the edges from Flow
        initialization and assign them explicitly.

        Args:
            - data (dict): the deserialized data

        Returns:
            - Flow

        """
        data["validate"] = False
        flow = super().create_object(data)
        return flow



from typing import TYPE_CHECKING, Any, Dict, Iterable

import prefect
import prefect.engine.signals
import prefect.triggers
from prefect.core.task import Task
from prefect.engine.results import PrefectResult

if TYPE_CHECKING:
    from prefect.core.flow import Flow  # pylint: disable=W0611


# A sentinel value indicating no default was provided
no_default = type(
    "no_default",
    (object,),
    dict.fromkeys(["__repr__", "__reduce__"], lambda s: "no_default"),
)()


class Parameter(Task):
    """
    A Parameter is a special task that defines a required flow input.

    A parameter's "slug" is automatically -- and immutably -- set to the parameter name.
    Flows enforce slug uniqueness across all tasks, so this ensures that the flow has
    no other parameters by the same name.

    Args:
        - name (str): the Parameter name.
        - default (any, optional): A default value for the parameter.
        - required (bool, optional): If True, the Parameter is required and the
            default value is ignored. Defaults to `False` if a `default` is
            provided, otherwise `True`.
        - tags ([str], optional): A list of tags for this parameter

    """

    def __init__(
        self,
        name: str,
        default: Any = no_default,
        required: bool = None,
        tags: Iterable[str] = None,
    ):
        if required is None:
            required = default is no_default
        if default is no_default:
            default = None
        self.required = required
        self.default = default

        super().__init__(
            name=name, slug=name, tags=tags, result=PrefectResult(), checkpoint=True,
        )

    def __repr__(self) -> str:
        return "<Parameter: {self.name}>".format(self=self)

    def __call__(self, flow: "Flow" = None) -> "Parameter":  # type: ignore
        """
        Calling a Parameter adds it to a flow.

        Args:
            - flow (Flow, optional): The flow to set dependencies on, defaults to the current
                flow in context if no flow is specified

        Returns:
            - Task: a new Task instance

        """
        result = super().bind(flow=flow)
        assert isinstance(result, Parameter)  # mypy assert
        return result

    def copy(self, name: str, **task_args: Any) -> "Task":  # type: ignore
        """
        Creates a copy of the Parameter with a new name.

        Args:
            - name (str): the new Parameter name
            - **task_args (dict, optional): a dictionary of task attribute keyword arguments,
                these attributes will be set on the new copy

        Raises:
            - AttributeError: if any passed `task_args` are not attributes of the original

        Returns:
            - Parameter: a copy of the current Parameter, with a new name and any attributes
                updated from `task_args`
        """
        return super().copy(name=name, slug=name, **task_args)

    def run(self) -> Any:
        params = prefect.context.get("parameters") or {}
        if self.required and self.name not in params:
            self.logger.debug(
                'Parameter "{}" was required but not provided.'.format(self.name)
            )
            raise prefect.engine.signals.FAIL(
                'Parameter "{}" was required but not provided.'.format(self.name)
            )
        return params.get(self.name, self.default)

    # Serialization ------------------------------------------------------------

    def serialize(self) -> Dict[str, Any]:
        """
        Creates a serialized representation of this parameter

        Returns:
            - dict representing this parameter
        """
        return prefect.serialization.task.ParameterSchema().dump(self)



"""
Collection of tasks for sending notifications.

Useful for situations in which state handlers are inappropriate.
"""
from prefect.tasks.notifications.email_task import EmailTask
from prefect.tasks.notifications.slack_task import SlackTask
from prefect.tasks.notifications.pushbullet_task import PushbulletTask



import json
from typing import Any

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class CreateGitHubPR(Task):
    """
    Task for opening/creating new GitHub Pull Requests using the v3 version of the GitHub REST API.

    Args:
        - repo (str, optional): the name of the repository to open the issue in; must be
            provided in the form `organization/repo_name`; can also be provided to the `run`
            method
        - title (str, optional): the title of the issue to create; can also be provided to the
            `run` method
        - body (str, optional): the contents of the issue; can also be provided to the `run` method
        - head (str, optional): the name of the branch where your changes are implemented; can also
            be provided to the `run` method
        - base (str, optional): the name of the branch you want the changes pulled into; can also
            be provided to the `run` method
        - **kwargs (Any, optional): additional keyword arguments to pass to the standard Task
            init method
    """

    def __init__(
        self,
        repo: str = None,
        title: str = None,
        body: str = None,
        head: str = None,
        base: str = None,
        **kwargs: Any
    ):
        self.repo = repo
        self.title = title
        self.body = body
        self.head = head
        self.base = base
        super().__init__(**kwargs)

    @defaults_from_attrs("repo", "title", "body", "head", "base")
    def run(
        self,
        repo: str = None,
        title: str = None,
        body: str = None,
        head: str = None,
        base: str = None,
        token: str = None,
    ) -> None:
        """
        Run method for this Task. Invoked by calling this Task after initialization within a
        Flow context, or by using `Task.bind`.

        Args:
            - repo (str, optional): the name of the repository to open the issue in; must be
                provided in the form `organization/repo_name`; defaults to the one provided at
                initialization
            - title (str, optional): the title of the issue to create; defaults to the one
                provided at initialization
            - body (str, optional): the contents of the issue; defaults to the one provided at
                initialization
            - head (str, optional): the name of the branch where your changes are implemented;
                defaults to the one provided at initialization
            - base (str, optional): the name of the branch you want the changes pulled into;
                defaults to the one provided at initialization
            - token (str): a GitHub API token

        Raises:
            - ValueError: if a `repo` was never provided
            - HTTPError: if the POST request returns a non-200 status code

        Returns:
            - None
        """
        if repo is None:
            raise ValueError("A GitHub repository must be provided.")

        # 'import requests' is expensive time-wise, we should do this just-in-time to keep
        # the 'import prefect' time low
        import requests

        url = "https://api.github.com/repos/{}/pulls".format(repo)
        headers = {
            "AUTHORIZATION": "token {}".format(token),
            "Accept": "application/vnd.github.v3+json",
        }
        pr = {"title": title, "body": body, "head": head, "base": base}

        # send the request
        resp = requests.post(url, data=json.dumps(pr), headers=headers)
        resp.raise_for_status()



import copy
import itertools
import multiprocessing
import os
import signal
import subprocess
import sys
import threading
import warnings
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError as FutureTimeout
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Union

import prefect

if TYPE_CHECKING:
    import prefect.engine.runner
    import prefect.engine.state
    from prefect.core.edge import Edge
    from prefect.core.task import Task
    from prefect.engine.state import State

StateList = Union["State", List["State"]]


def run_with_heartbeat(
    runner_method: Callable[..., "prefect.engine.state.State"]
) -> Callable[..., "prefect.engine.state.State"]:
    """
    Utility decorator for running class methods with a heartbeat.  The class should implement
    `self._heartbeat` with no arguments.
    """

    @wraps(runner_method)
    def inner(
        self: "prefect.engine.runner.Runner", *args: Any, **kwargs: Any
    ) -> "prefect.engine.state.State":
        try:
            p = None
            try:
                if self._heartbeat():
                    # we use Popen + a prefect CLI for a few reasons:
                    # - using threads would interfere with the task; for example, a task
                    #   which does not release the GIL would prevent the heartbeat thread from
                    #   firing
                    # - using multiprocessing.Process would release the GIL but a subprocess
                    #   cannot be spawned from a deamonic subprocess, and Dask sometimes will
                    #   submit tasks to run within daemonic subprocesses
                    current_env = dict(os.environ).copy()
                    auth_token = prefect.context.config.cloud.get("auth_token")
                    api_url = prefect.context.config.cloud.get("api")
                    current_env.setdefault("PREFECT__CLOUD__AUTH_TOKEN", auth_token)
                    current_env.setdefault("PREFECT__CLOUD__API", api_url)
                    clean_env = {k: v for k, v in current_env.items() if v is not None}
                    p = subprocess.Popen(
                        self.heartbeat_cmd,
                        env=clean_env,
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )
            except Exception:
                self.logger.exception(
                    "Heartbeat failed to start.  This could result in a zombie run."
                )
            return runner_method(self, *args, **kwargs)
        finally:
            if p is not None:
                exit_code = p.poll()
                if exit_code is not None:
                    msg = "Heartbeat process died with exit code {}".format(exit_code)
                    self.logger.error(msg)
                p.kill()

    return inner


def main_thread_timeout(
    fn: Callable, *args: Any, timeout: int = None, **kwargs: Any
) -> Any:
    """
    Helper function for implementing timeouts on function executions.
    Implemented by setting a `signal` alarm on a timer. Must be run in the main thread.
    Args:
        - fn (callable): the function to execute
        - *args (Any): arguments to pass to the function
        - timeout (int): the length of time to allow for
            execution before raising a `TimeoutError`, represented as an integer in seconds
        - **kwargs (Any): keyword arguments to pass to the function
    Returns:
        - the result of `f(*args, **kwargs)`
    Raises:
        - TimeoutError: if function execution exceeds the allowed timeout
        - ValueError: if run from outside the main thread
    """

    if timeout is None:
        return fn(*args, **kwargs)

    def error_handler(signum, frame):  # type: ignore
        raise TimeoutError("Execution timed out.")

    try:
        signal.signal(signal.SIGALRM, error_handler)
        signal.alarm(timeout)
        return fn(*args, **kwargs)
    finally:
        signal.alarm(0)


def multiprocessing_timeout(
    fn: Callable, *args: Any, timeout: int = None, **kwargs: Any
) -> Any:
    """
    Helper function for implementing timeouts on function executions.
    Implemented by spawning a new multiprocess.Process() and joining with timeout.
    Args:
        - fn (callable): the function to execute
        - *args (Any): arguments to pass to the function
        - timeout (int): the length of time to allow for
            execution before raising a `TimeoutError`, represented as an integer in seconds
        - **kwargs (Any): keyword arguments to pass to the function
    Returns:
        - the result of `f(*args, **kwargs)`
    Raises:
        - AssertionError: if run from a daemonic process
        - TimeoutError: if function execution exceeds the allowed timeout
    """

    if timeout is None:
        return fn(*args, **kwargs)

    def retrieve_value(
        *args: Any, _container: multiprocessing.Queue, _ctx_dict: dict, **kwargs: Any
    ) -> None:
        """Puts the return value in a multiprocessing-safe container"""
        try:
            with prefect.context(_ctx_dict):
                val = fn(*args, **kwargs)
            _container.put(val)
        except Exception as exc:
            _container.put(exc)

    q = multiprocessing.Queue()  # type: multiprocessing.Queue
    kwargs["_container"] = q
    kwargs["_ctx_dict"] = prefect.context.to_dict()
    p = multiprocessing.Process(target=retrieve_value, args=args, kwargs=kwargs)
    p.start()
    p.join(timeout)
    p.terminate()
    if not q.empty():
        res = q.get()
        if isinstance(res, Exception):
            raise res
        return res
    else:
        raise TimeoutError("Execution timed out.")


def timeout_handler(
    fn: Callable, *args: Any, timeout: int = None, **kwargs: Any
) -> Any:
    """
    Helper function for implementing timeouts on function executions.

    The exact implementation varies depending on whether this function is being
    run in the main thread or a non-daemonic subprocess.  If this is run from a
    daemonic subprocess or on Windows, the task is run in a
    `ThreadPoolExecutor` and only a soft timeout is enforced, meaning a
    `TimeoutError` is raised at the appropriate time but the task continues
    running in the background.

    Args:
        - fn (callable): the function to execute
        - *args (Any): arguments to pass to the function
        - timeout (int): the length of time to allow for
            execution before raising a `TimeoutError`, represented as an integer in seconds
        - **kwargs (Any): keyword arguments to pass to the function

    Returns:
        - the result of `f(*args, **kwargs)`

    Raises:
        - TimeoutError: if function execution exceeds the allowed timeout
    """
    # if no timeout, just run the function
    if timeout is None:
        return fn(*args, **kwargs)

    # if we are running the main thread, use a signal to stop execution at the
    # appropriate time; else if we are running in a non-daemonic process, spawn
    # a subprocess to kill at the appropriate time
    if not sys.platform.startswith("win"):
        if threading.current_thread() is threading.main_thread():
            return main_thread_timeout(fn, *args, timeout=timeout, **kwargs)
        elif multiprocessing.current_process().daemon is False:
            return multiprocessing_timeout(fn, *args, timeout=timeout, **kwargs)

        msg = (
            "This task is running in a daemonic subprocess; "
            "consequently Prefect can only enforce a soft timeout limit, i.e., "
            "if your Task reaches its timeout limit it will enter a TimedOut state "
            "but continue running in the background."
        )
    else:
        msg = (
            "This task is running on Windows; "
            "consequently Prefect can only enforce a soft timeout limit, i.e., "
            "if your Task reaches its timeout limit it will enter a TimedOut state "
            "but continue running in the background."
        )

    warnings.warn(msg)
    executor = ThreadPoolExecutor()

    def run_with_ctx(*args: Any, _ctx_dict: dict, **kwargs: Any) -> Any:
        with prefect.context(_ctx_dict):
            return fn(*args, **kwargs)

    fut = executor.submit(
        run_with_ctx, *args, _ctx_dict=prefect.context.to_dict(), **kwargs
    )

    try:
        return fut.result(timeout=timeout)
    except FutureTimeout:
        raise TimeoutError("Execution timed out.")


class RecursiveCall(Exception):
    def __init__(self, func: Callable, *args: Any, **kwargs: Any):
        self.func = func
        self.args = args
        self.kwargs = kwargs


def tail_recursive(func: Callable) -> Callable:
    """
    Helper function to facilitate tail recursion of the wrapped function.

    This allows for recursion with unlimited depth since a stack is not allocated for
    each "nested" call. Note: instead of calling the target function in question, a
    `RecursiveCall` exception must be raised instead.

    Args:
        - fn (callable): the function to execute

    Returns:
        - the result of `f(*args, **kwargs)`

    Raises:
        - RecursionError: if a recursive "call" (raised exception) is made with a function that is
            not decorated with `tail_recursive` decorator.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        while True:
            try:
                return func(*args, **kwargs)
            except RecursiveCall as exc:
                try:
                    call_func = getattr(exc.func, "__wrapped_func__")
                except AttributeError:
                    raise RecursionError(
                        "function has not been wrapped to provide tail recursion (func={})".format(
                            exc.func
                        )
                    )

                # there may be multiple nested recursive calls, we should only
                # respond to calls for the wrapped function explicitly,
                # otherwise allow the call to continue to propagate
                if call_func != func:
                    raise exc
                args = exc.args
                kwargs = exc.kwargs
                continue

    setattr(wrapper, "__wrapped_func__", func)
    return wrapper


def prepare_upstream_states_for_mapping(
    state: "State",
    upstream_states: "Dict[Edge, State]",
    mapped_children: "Dict[Task, list]",
    executor: "prefect.engine.executors.Executor",
) -> list:
    """
    If the task is being mapped, submits children tasks for execution. Returns a `Mapped` state.

    Args:
        - state (State): the parent task's current state
        - upstream_states (Dict[Edge, State]): the upstream states to this task
        - mapped_children (Dict[Task, List[State]]): any mapped children upstream of this task

    Returns:
        - List: a restructured list of upstream states correponding to each new mapped child task
    """

    # if the current state is failed / skipped or otherwise
    # in a state that signifies we should not continue with mapping,
    # we return an empty list
    if state.is_pending() or state.is_failed() or state.is_skipped():
        return []

    map_upstream_states = []

    # copy the mapped children dict to avoid mutating the global one
    # when we flatten mapped children
    mapped_children = mapped_children.copy()

    # we don't know how long the iterables are, but we want to iterate until we reach
    # the end of the shortest one
    counter = itertools.count()

    # preprocessing
    for edge, upstream_state in upstream_states.items():

        # ensure we are working with populated result objects
        if edge.key in state.cached_inputs:
            upstream_state._result = state.cached_inputs[edge.key]

        # if the upstream was mapped and the edge is flattened, we need
        # to process the mapped children (which could be futures) into a
        # flat structure
        if upstream_state.is_mapped() and edge.flattened:
            mapped_children[edge.upstream_task] = flatten_mapped_children(
                mapped_children=mapped_children[edge.upstream_task], executor=executor
            )

    # infinite loop, if upstream_states has any entries
    while True and upstream_states:
        i = next(counter)
        states = {}
        try:

            for edge, upstream_state in upstream_states.items():

                # if the edge is not mapped over, then we take its state
                if not edge.mapped:
                    states[edge] = upstream_state

                # if the edge is mapped and the upstream state is Mapped, then we are mapping
                # over a mapped task. In this case, we take the appropriately-indexed upstream
                # state from the upstream tasks's `Mapped.map_states` array.
                # Note that these "states" might actually be futures at this time; we aren't
                # blocking until they finish.
                elif edge.mapped and upstream_state.is_mapped():
                    states[edge] = mapped_children[edge.upstream_task][i]  # type: ignore

                # Otherwise, we are mapping over the result of a "vanilla" task. In this
                # case, we create a copy of the upstream state but set the result to the
                # appropriately-indexed item from the upstream task's `State.result`
                # array.
                else:
                    states[edge] = copy.copy(upstream_state)

                    # if the current state is already Mapped, then we might be executing
                    # a re-run of the mapping pipeline. In that case, the upstream states
                    # might not have `result` attributes.
                    # Therefore, we only try to get a result if EITHER this task's
                    # state is not already mapped OR the upstream result is not None.
                    if (
                        not state.is_mapped()
                        or upstream_state._result != prefect.engine.result.NoResult
                    ):
                        # this line should never be hit due to a check
                        # in the TaskRunner when evaluating the mapped parent
                        if not hasattr(upstream_state.result, "__getitem__"):
                            raise TypeError(
                                (
                                    "Cannot map over unsubscriptable object of type {t}: {val}..."
                                ).format(
                                    t=type(upstream_state.result),
                                    val=repr(upstream_state.result)[:10],
                                )
                            )
                        upstream_result = upstream_state._result.from_value(  # type: ignore
                            upstream_state.result[i]
                        )
                        states[edge].result = upstream_result
                    elif state.is_mapped():
                        if i >= len(state.map_states):  # type: ignore
                            raise IndexError()

            # only add this iteration if we made it through all iterables
            map_upstream_states.append(states)

        # index error means we reached the end of the shortest iterable
        except IndexError:
            break

    return map_upstream_states


def _build_flattened_state(state: "State", index: int) -> "State":
    """Helper function for `flatten_upstream_state`"""
    new_state = copy.copy(state)
    new_state.result = state._result.from_value(  # type: ignore
        state.result[index]
    )
    return new_state


def flatten_upstream_state(upstream_state: "State") -> "State":
    """
    Given an upstream state, returns its result as a flattened list. If
    flattening fails, the object is returned unmodified.
    """
    try:
        # attempt to unnest
        flattened_result = [y for x in upstream_state.result for y in x]
    except TypeError:
        return upstream_state

    new_state = copy.copy(upstream_state)
    new_state.result = new_state._result.from_value(  # type: ignore
        flattened_result
    )
    return new_state


def flatten_mapped_children(
    mapped_children: List["State"], executor: "prefect.engine.executors.Executor",
) -> List["State"]:
    counts = executor.wait(
        [executor.submit(lambda c: len(c._result.value), c) for c in mapped_children]
    )
    new_states = []

    for child, count in zip(mapped_children, counts):
        new_states.append(
            [executor.submit(_build_flattened_state, child, i) for i in range(count)]
        )

    flattened_states = [i for s in new_states for i in s]
    return flattened_states



"""
A state-handler that will create and assign a Jira ticket.
"""

from typing import TYPE_CHECKING, Union, cast, Optional
from datetime import datetime
from toolz import curry

import prefect


try:
    from jira import JIRA
except ImportError:
    raise ImportError(
        'Using `jira_notifier` requires Prefect to be installed with the "jira" extra.'
    )

if TYPE_CHECKING:
    import prefect.engine.state
    import prefect.client
    from prefect import Flow, Task


def jira_message_formatter(
    tracked_obj: "Union[Flow, Task]", state: "prefect.engine.state.State"
) -> str:
    time = datetime.utcnow()
    msg = "Message from Prefect: {0} entered a {1} state at {2}".format(
        tracked_obj.name, type(state).__name__, time
    )
    return msg


@curry
def jira_notifier(
    tracked_obj: "Union[Flow, Task]",
    old_state: "prefect.engine.state.State",
    new_state: "prefect.engine.state.State",
    ignore_states: list = None,
    only_states: list = None,
    server_URL: str = None,
    options: Optional[dict] = None,
    assignee: str = "-1",
) -> "prefect.engine.state.State":
    """
    Jira Notifier requires a Jira account and API token.  The API token can be created at:
    https://id.atlassian.com/manage/api-tokens The Jira account username ('JIRAUSER'), API
    token ('JIRATOKEN') should be set as part of a 'JIRASECRETS' object in Prefect Secrets.

    An example 'JIRASECRETS' secret configuration looks like:

    ```toml
    [secrets]
    JIRASECRETS.JIRATOKEN = "XXXXXXXXX"
    JIRASECRETS.JIRAUSER = "xxxxx@yyy.com"
    JIRASECRETS.JIRASERVER = "https://???.atlassian.net"
    ```

    The server URL can be set as part of the 'JIRASECRETS' object ('JIRASERVER') or passed to
    the jira notifier state handler as the "server_URL" argument.  Jira Notifier works as a
    standalone state handler, or can be called from within a custom state handler.  This
    function is curried meaning that it can be called multiple times to partially bind any
    keyword arguments (see example below).  Jira Notifier creates a new ticket with the
    information about the task or flow it is bound to when that task or flow is in a specific
    state.  (For example it will create a ticket to tell you that the flow you set it on is in
    a failed state.) You should use the options dictionary to set the project name and issue
    type.  You can use the "assignee" argument to assign that ticket to a specific member of
    your team.

    Args:
        - tracked_obj (Task or Flow): Task or Flow object the handler is registered with
        - old_state (State): previous state of tracked object
        - new_state (State): new state of tracked object
        - options (Dictionary): Must inlucde a 'project' key and an 'issuetype' key (e.g.
            options = {'project': 'TEST', 'issuetype': {'name': 'task'}}). For jira service
            desk tickets, the issue type should use the request type id e.g. 'issuetype':
            {'id': '10010'}. A description can be added using the key 'description'. Custom
            fields can also be added e.g.  'customfield_10017': 'SDTS/flowdown'
        - ignore_states ([State], optional): list of `State` classes to ignore, e.g.,
            `[Running, Scheduled]`. If `new_state` is an instance of one of the passed states,
            no notification will occur.
        - only_states ([State], optional): similar to `ignore_states`, but instead _only_
            notifies you if the Task / Flow is in a state from the provided list of `State`
            classes
        - server_URL (String): The URL of your atlassian account e.g.
            "https://test.atlassian.net".  Can also be set as a Prefect Secret.
        - assignee: the atlassian username of the person you want to assign the ticket to.
            Defaults to "automatic" if this is not set.

    Returns:
        - State: the `new_state` object that was provided

    Raises:
        - ValueError: if the jira ticket creation or assignment fails for any reason

    Example:
        ```python
        from prefect import task
        from prefect.utilities.jira_notification import jira_notifier

        @task(state_handlers=[
            jira_notifier(
                only_states=[Failed],
                options={'project': 'TEST', 'issuetype': {'name': 'Task'}},
                assignee='tester'
            )
        ])
        def add(x, y):
            return x + y
        ```
    """

    options = options or dict()
    jira_credentials = cast(dict, prefect.client.Secret("JIRASECRETS").get())
    username = jira_credentials["JIRAUSER"]
    password = jira_credentials["JIRATOKEN"]

    if not server_URL:
        server_URL = jira_credentials["JIRASERVER"]

    ignore_states = ignore_states or []
    only_states = only_states or []

    if any([isinstance(new_state, ignored) for ignored in ignore_states]):
        return new_state

    if only_states and not any(
        [isinstance(new_state, included) for included in only_states]
    ):
        return new_state

    summary_text = str(jira_message_formatter(tracked_obj, new_state))

    options["summary"] = summary_text

    jira = JIRA(basic_auth=(username, password), options={"server": server_URL})

    created = jira.create_issue(options)
    if not created:
        raise ValueError("Creating Jira Issue for {} failed".format(tracked_obj))

    if assignee:
        assigned = jira.assign_issue(created, assignee)
        if not assigned:
            raise ValueError("Assigning Jira issue for {} failed".format(tracked_obj))

    return new_state



from typing import TYPE_CHECKING, Any, Dict, List


from prefect.environments.storage import Storage
from prefect.utilities.storage import extract_flow_from_file

if TYPE_CHECKING:
    from prefect.core.flow import Flow


class GitHub(Storage):
    """
    GitHub storage class. This class represents the Storage interface for Flows stored
    in `.py` files in a GitHub repository.

    This class represents a mapping of flow name to file paths contained in the git repo,
    meaning that all flow files should be pushed independently. A typical workflow using
    this storage type might look like the following:

    - Compose flow `.py` file where flow has GitHub storage:

    ```python
    flow = Flow("my-flow")
    flow.storage = GitHub(repo="my/repo", path="/flows/flow.py")
    ```

    - Push this `flow.py` file to the `my/repo` repository under `/flows/flow.py`.

    - Call `prefect register -f flow.py` to register this flow with GitHub storage.

    Args:
        - repo (str): the name of a GitHub repository to store this Flow
        - path (str, optional): a path pointing to a flow file in the repo
        - **kwargs (Any, optional): any additional `Storage` initialization options
    """

    def __init__(self, repo: str, path: str = None, **kwargs: Any) -> None:
        self.flows = dict()  # type: Dict[str, str]
        self._flows = dict()  # type: Dict[str, "Flow"]
        self.repo = repo
        self.path = path

        super().__init__(**kwargs)

    @property
    def default_labels(self) -> List[str]:
        return ["github-flow-storage"]

    def get_flow(self, flow_location: str = None) -> "Flow":
        """
        Given a flow_location within this Storage object, returns the underlying Flow (if possible).
        If the Flow is not found an error will be logged and `None` will be returned.

        Args:
            - flow_location (str): the location of a flow within this Storage; in this case,
                a file path on a repository where a Flow file has been committed. Will use `path` if not
                provided.

        Returns:
            - Flow: the requested Flow

        Raises:
            - ValueError: if the flow is not contained in this storage
            - UnknownObjectException: if the flow file is unable to be retrieved
        """
        if flow_location:
            if flow_location not in self.flows.values():
                raise ValueError("Flow is not contained in this Storage")
        elif self.path:
            flow_location = self.path
        else:
            raise ValueError("No flow location provided")

        from github import UnknownObjectException

        repo = self._github_client.get_repo(self.repo)

        try:
            contents = repo.get_contents(flow_location)
            decoded_contents = contents.decoded_content
        except UnknownObjectException as exc:
            self.logger.error(
                "Error retrieving file contents from {} on repo {}. Ensure the file exists.".format(
                    flow_location, self.repo
                )
            )
            raise exc

        return extract_flow_from_file(file_contents=decoded_contents)

    def add_flow(self, flow: "Flow") -> str:
        """
        Method for storing a new flow as bytes in the local filesytem.

        Args:
            - flow (Flow): a Prefect Flow to add

        Returns:
            - str: the location of the added flow in the repo

        Raises:
            - ValueError: if a flow with the same name is already contained in this storage
        """
        if flow.name in self:
            raise ValueError(
                'Name conflict: Flow with the name "{}" is already present in this storage.'.format(
                    flow.name
                )
            )

        self.flows[flow.name] = self.path  # type: ignore
        self._flows[flow.name] = flow
        return self.path  # type: ignore

    def build(self) -> "Storage":
        """
        Build the GitHub storage object and run basic healthchecks. Due to this object
        supporting file based storage no files are committed to the repository during
        this step. Instead, all files should be committed independently.

        Returns:
            - Storage: a GitHub object that contains information about how and where
                each flow is stored
        """
        self.run_basic_healthchecks()

        return self

    def __contains__(self, obj: Any) -> bool:
        """
        Method for determining whether an object is contained within this storage.
        """
        if not isinstance(obj, str):
            return False
        return obj in self.flows

    @property
    def _github_client(self):  # type: ignore
        from prefect.utilities.git import get_github_client

        return get_github_client()
